# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT

from airflow.providers.google.cloud.operators.dataform import (
    DataformCancelWorkflowInvocationOperator,
    DataformCreateCompilationResultOperator,
    DataformCreateRepositoryOperator,
    DataformCreateWorkflowInvocationOperator,
    DataformCreateWorkspaceOperator,
    DataformDeleteRepositoryOperator,
    DataformDeleteWorkspaceOperator,
    DataformGetCompilationResultOperator,
    DataformGetWorkflowInvocationOperator,
    DataformInstallNpmPackagesOperator,
    DataformMakeDirectoryOperator,
    DataformQueryWorkflowInvocationActionsOperator,
    DataformRemoveDirectoryOperator,
    DataformRemoveFileOperator,
    DataformWriteFileOperator,
)

HOOK_STR = "airflow.providers.google.cloud.operators.dataform.DataformHook"
WORKFLOW_INVOCATION_STR = "airflow.providers.google.cloud.operators.dataform.WorkflowInvocation"
WORKFLOW_INVOCATION_ACTION_STR = "airflow.providers.google.cloud.operators.dataform.WorkflowInvocationAction"
COMPILATION_RESULT_STR = "airflow.providers.google.cloud.operators.dataform.CompilationResult"
REPOSITORY_STR = "airflow.providers.google.cloud.operators.dataform.Repository"
WORKSPACE_STR = "airflow.providers.google.cloud.operators.dataform.Workspace"
WRITE_FILE_RESPONSE_STR = "airflow.providers.google.cloud.operators.dataform.WriteFileResponse"
MAKE_DIRECTORY_RESPONSE_STR = "airflow.providers.google.cloud.operators.dataform.MakeDirectoryResponse"
INSTALL_NPM_PACKAGES_RESPONSE_STR = (
    "airflow.providers.google.cloud.operators.dataform.InstallNpmPackagesResponse"
)

PROJECT_ID = "project-id"
REGION = "region"
REPOSITORY_ID = "test_repository_id"
WORKSPACE_ID = "test_workspace_id"
WORKSPACE = f"projects/{PROJECT_ID}/locations/{REGION}/repositories/{REPOSITORY_ID}/workspaces/{WORKSPACE_ID}"
COMPILATION_RESULT_ID = "test_compilation_result_id"
GCP_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
FILEPATH = "path/to/file.txt"
FILE_CONTENT = b"test content"
DIRECTORY_PATH = "path/to/directory"

WORKFLOW_INVOCATION = {
    "compilation_result": (
        f"projects/{PROJECT_ID}/locations/{REGION}/repositories/{REPOSITORY_ID}/"
        f"compilationResults/{COMPILATION_RESULT_ID}"
    )
}
WORKFLOW_INVOCATION_ID = "test_workflow_invocation"


class TestDataformCreateCompilationResult:
    @mock.patch(HOOK_STR)
    @mock.patch(COMPILATION_RESULT_STR)
    def test_execute(self, compilation_result_mock, hook_mock):
        compilation_result = {
            "git_commitish": "main",
            "workspace": WORKSPACE,
        }
        op = DataformCreateCompilationResultOperator(
            task_id="create_compilation_result",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            compilation_result=compilation_result,
        )
        compilation_result_mock.return_value.to_dict.return_value = None
        op.execute(context=mock.MagicMock())
        hook_mock.return_value.create_compilation_result.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            compilation_result=compilation_result,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformGetCompilationResultOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(COMPILATION_RESULT_STR)
    def test_execute(self, compilation_result_mock, hook_mock):
        op = DataformGetCompilationResultOperator(
            task_id="get_compilation_result",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            compilation_result_id=COMPILATION_RESULT_ID,
        )
        compilation_result_mock.return_value.to_dict.return_value = None
        op.execute(context=mock.MagicMock())
        hook_mock.return_value.get_compilation_result.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            compilation_result_id=COMPILATION_RESULT_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformCreateWorkflowInvocationOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(WORKFLOW_INVOCATION_STR)
    def test_execute(self, workflow_invocation_str, hook_mock):
        op = DataformCreateWorkflowInvocationOperator(
            task_id="create_workflow_invocation",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation=WORKFLOW_INVOCATION,
        )
        workflow_invocation_str.return_value.to_dict.return_value = None
        op.execute(context=mock.MagicMock())

        hook_mock.return_value.create_workflow_invocation.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation=WORKFLOW_INVOCATION,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformGetWorkflowInvocationOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(WORKFLOW_INVOCATION_STR)
    def test_execute(self, workflow_invocation_str, hook_mock):
        op = DataformGetWorkflowInvocationOperator(
            task_id="get_workflow_invocation",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
        )

        workflow_invocation_str.return_value.to_dict.return_value = None
        op.execute(context=mock.MagicMock())

        hook_mock.return_value.get_workflow_invocation.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformQueryWorkflowInvocationActionsOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(WORKFLOW_INVOCATION_ACTION_STR)
    def test_execute(self, workflow_invocation_action_str, hook_mock):
        op = DataformQueryWorkflowInvocationActionsOperator(
            task_id="query_workflow_invocation_action",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
        )

        workflow_invocation_action_str.return_value.to_dict.return_value = None
        op.execute(context=mock.MagicMock())

        hook_mock.return_value.query_workflow_invocation_actions.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformCancelWorkflowInvocationOperator:
    @mock.patch(HOOK_STR)
    def test_execute(self, hook_mock):
        op = DataformCancelWorkflowInvocationOperator(
            task_id="cancel_workflow_invocation",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.return_value.cancel_workflow_invocation.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workflow_invocation_id=WORKFLOW_INVOCATION_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformCreateRepositoryOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(REPOSITORY_STR)
    def test_execute(self, _, hook_mock):
        op = DataformCreateRepositoryOperator(
            task_id="create-repository",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.return_value.create_repository.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformDeleteRepositoryOperator:
    @mock.patch(HOOK_STR)
    def test_execute(self, hook_mock):
        force = True
        op = DataformDeleteRepositoryOperator(
            task_id="delete-repository",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            force=force,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.delete_repository.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            force=force,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformCreateWorkspaceOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(WORKSPACE_STR)
    def test_execute(self, _, hook_mock):
        op = DataformCreateWorkspaceOperator(
            task_id="create-workspace",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.create_workspace.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformDeleteWorkspaceOperator:
    @mock.patch(HOOK_STR)
    def test_execute(self, hook_mock):
        op = DataformDeleteWorkspaceOperator(
            task_id="delete-workspace",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.delete_workspace.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformWriteFileOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(WRITE_FILE_RESPONSE_STR)
    def test_execute(self, _, hook_mock):
        op = DataformWriteFileOperator(
            task_id="write-file",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            filepath=FILEPATH,
            contents=FILE_CONTENT,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.write_file.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            filepath=FILEPATH,
            contents=FILE_CONTENT,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformMakeDirectoryOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(MAKE_DIRECTORY_RESPONSE_STR)
    def test_execute(self, _, hook_mock):
        op = DataformMakeDirectoryOperator(
            task_id="write-file",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            directory_path=DIRECTORY_PATH,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.make_directory.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            path=DIRECTORY_PATH,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformRemoveFileOperator:
    @mock.patch(HOOK_STR)
    def test_execute(self, hook_mock):
        op = DataformRemoveFileOperator(
            task_id="remove-file",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            filepath=FILEPATH,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.remove_file.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            filepath=FILEPATH,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformRemoveDirectoryOperator:
    @mock.patch(HOOK_STR)
    def test_execute(self, hook_mock):
        op = DataformRemoveDirectoryOperator(
            task_id="remove-directory",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            directory_path=DIRECTORY_PATH,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.remove_directory.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            path=DIRECTORY_PATH,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestDataformInstallNpmPackagesOperator:
    @mock.patch(HOOK_STR)
    @mock.patch(INSTALL_NPM_PACKAGES_RESPONSE_STR)
    def test_execute(self, _, hook_mock):
        op = DataformInstallNpmPackagesOperator(
            task_id="remove-directory",
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
        )

        op.execute(context=mock.MagicMock())
        hook_mock.return_value.install_npm_packages.assert_called_once_with(
            project_id=PROJECT_ID,
            region=REGION,
            repository_id=REPOSITORY_ID,
            workspace_id=WORKSPACE_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for creating GKE cluster with Ray enabled.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.kubernetes_engine import (
    GKECreateClusterOperator,
    GKEDeleteClusterOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "kubernetes_engine_ray"
GCP_PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

GCP_LOCATION = "europe-west1"
CLUSTER_NAME_BASE = f"cluster-{DAG_ID}".replace("_", "-")
CLUSTER_NAME_FULL = CLUSTER_NAME_BASE + f"-{ENV_ID}".replace("_", "-")
CLUSTER_NAME = CLUSTER_NAME_BASE if len(CLUSTER_NAME_FULL) >= 33 else CLUSTER_NAME_FULL

# [START howto_operator_gcp_gke_create_cluster_definition_with_ray]
CLUSTER = {
    "name": CLUSTER_NAME,
    "node_pools": [
        {
            "name": f"{CLUSTER_NAME}-node",
            "initial_node_count": 1,
        },
    ],
    "autopilot": {"enabled": True},
    "addons_config": {
        "ray_operator_config": {"enabled": True},
    },
}
# [END howto_operator_gcp_gke_create_cluster_definition_with_ray]


with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    create_cluster = GKECreateClusterOperator(
        task_id="create_cluster",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        body=CLUSTER,
    )

    delete_cluster = GKEDeleteClusterOperator(
        task_id="delete_cluster",
        cluster_name=CLUSTER_NAME,
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
    )
    delete_cluster.trigger_rule = TriggerRule.ALL_DONE

    create_cluster >> delete_cluster

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define, field

from airflow.providers.google import __version__ as provider_version

if TYPE_CHECKING:
    from openlineage.client.generated.base import JobFacet, RunFacet

try:
    try:
        from openlineage.client.generated.base import RunFacet
    except ImportError:  # Old OpenLineage client is used
        from openlineage.client.facet import BaseFacet as RunFacet  # type: ignore[assignment]
    try:
        from openlineage.client.generated.base import JobFacet
    except ImportError:  # Old OpenLineage client is used
        from openlineage.client.facet import BaseFacet as JobFacet  # type: ignore[assignment]

    @define
    class BigQueryJobRunFacet(RunFacet):
        """
        Facet that represents relevant statistics of bigquery run.

        :param cached: BigQuery caches query results. Rest of the statistics will not be provided for cached queries.
        :param billedBytes: How many bytes BigQuery bills for.
        :param properties: Full property tree of BigQUery run.
        """

        cached: bool
        billedBytes: int | None = field(default=None)
        properties: str | None = field(default=None)

        @staticmethod
        def _get_schema() -> str:
            return (
                "https://raw.githubusercontent.com/apache/airflow/"
                f"providers-google/{provider_version}/airflow/providers/google/"
                "openlineage/BigQueryJobRunFacet.json"
            )

    @define
    class CloudStorageTransferJobFacet(JobFacet):
        """
        Facet representing a Cloud Storage Transfer Service job configuration.

        :param jobName: Unique name of the transfer job.
        :param projectId: GCP project where the transfer job is defined.
        :param description: User-provided description of the transfer job.
        :param status: Current status of the transfer job (e.g. "ENABLED", "DISABLED").
        :param sourceBucket: Name of the source bucket (e.g. AWS S3).
        :param sourcePath: Prefix/path inside the source bucket.
        :param targetBucket: Name of the destination bucket (e.g. GCS).
        :param targetPath: Prefix/path inside the destination bucket.
        :param objectConditions: Object selection rules (e.g. include/exclude prefixes).
        :param transferOptions: Transfer options, such as overwrite behavior or whether to delete objects
        from the source after transfer.
        :param schedule: Schedule for the transfer job (if recurring).
        """

        jobName: str | None = field(default=None)
        projectId: str | None = field(default=None)
        description: str | None = field(default=None)
        status: str | None = field(default=None)
        sourceBucket: str | None = field(default=None)
        sourcePath: str | None = field(default=None)
        targetBucket: str | None = field(default=None)
        targetPath: str | None = field(default=None)
        objectConditions: dict | None = field(default=None)
        transferOptions: dict | None = field(default=None)
        schedule: dict | None = field(default=None)

        @staticmethod
        def _get_schema() -> str:
            return (
                "https://raw.githubusercontent.com/apache/airflow/"
                f"providers-google/{provider_version}/airflow/providers/google/"
                "openlineage/CloudStorageTransferJobFacet.json"
            )

    @define
    class CloudStorageTransferRunFacet(RunFacet):
        """
        Facet representing a Cloud Storage Transfer Service job execution run.

        :param jobName: Name of the transfer job being executed.
        :param operationName: Name of the specific transfer operation instance.
        :param status: Current status of the operation (e.g. "IN_PROGRESS", "SUCCESS", "FAILED").
        :param startTime: Time when the transfer job execution started (ISO 8601 format).
        :param endTime: Time when the transfer job execution finished (ISO 8601 format).
        :param wait: Whether the operator waits for the job to complete before finishing.
        :param timeout: Timeout (in seconds) for the transfer run to complete.
        :param deferrable: Whether the operator defers execution until job completion.
        :param deleteJobAfterCompletion: Whether the operator deletes the transfer job after the run completes.
        """

        jobName: str | None = field(default=None)
        operationName: str | None = field(default=None)
        status: str | None = field(default=None)
        startTime: str | None = field(default=None)
        endTime: str | None = field(default=None)
        wait: bool = field(default=True)
        timeout: float | None = field(default=None)
        deferrable: bool = field(default=False)
        deleteJobAfterCompletion: bool = field(default=False)

        @staticmethod
        def _get_schema() -> str:
            return (
                "https://raw.githubusercontent.com/apache/airflow/"
                f"providers-google/{provider_version}/airflow/providers/google/"
                "openlineage/CloudStorageTransferRunFacet.json"
            )

    @define
    class DataFusionRunFacet(RunFacet):
        """
        Facet that represents relevant details of a Cloud Data Fusion pipeline run.

        :param runId: The pipeline execution id.
        :param runtimeArgs: Runtime arguments passed to the pipeline.
        """

        runId: str | None = field(default=None)
        runtimeArgs: dict[str, str] | None = field(default=None)

        @staticmethod
        def _get_schema() -> str:
            return (
                "https://raw.githubusercontent.com/apache/airflow/"
                f"providers-google/{provider_version}/airflow/providers/google/"
                "openlineage/DataFusionRunFacet.json"
            )

except ImportError:  # OpenLineage is not available

    def create_no_op(*_, **__) -> None:
        """
        Create a no-op placeholder.

        This function creates and returns a None value, used as a placeholder when the OpenLineage client
        library is available. It represents an action that has no effect.
        """
        return None

    BigQueryJobRunFacet = create_no_op  # type: ignore[misc, assignment]
    CloudStorageTransferJobFacet = create_no_op  # type: ignore[misc, assignment]
    CloudStorageTransferRunFacet = create_no_op  # type: ignore[misc, assignment]
    DataFusionRunFacet = create_no_op  # type: ignore[misc, assignment]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that demonstrates operators for the Google Cloud Video Intelligence service in the Google
Cloud Platform.

This DAG relies on the following OS environment variables:

* BUCKET_NAME - Google Cloud Storage bucket where the file exists.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.api_core.retry import Retry

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.operators.video_intelligence import (
    CloudVideoIntelligenceDetectVideoExplicitContentOperator,
    CloudVideoIntelligenceDetectVideoLabelsOperator,
    CloudVideoIntelligenceDetectVideoShotsOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator
from airflow.providers.standard.operators.bash import BashOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")

DAG_ID = "example_gcp_video_intelligence"

# Public bucket holding the sample data
BUCKET_NAME_SRC = "cloud-samples-data"
# Path to the data inside the public bucket
PATH_SRC = "video/cat.mp4"

# [START howto_operator_video_intelligence_os_args]
BUCKET_NAME_DST = f"bucket-src-{DAG_ID}-{ENV_ID}"
# [END howto_operator_video_intelligence_os_args]

FILE_NAME = "video.mp4"

# [START howto_operator_video_intelligence_other_args]
INPUT_URI = f"gs://{BUCKET_NAME_DST}/{FILE_NAME}"
# [END howto_operator_video_intelligence_other_args]

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME_DST)

    copy_single_file = GCSToGCSOperator(
        task_id="copy_single_gcs_file",
        source_bucket=BUCKET_NAME_SRC,
        source_object=PATH_SRC,
        destination_bucket=BUCKET_NAME_DST,
        destination_object=FILE_NAME,
    )

    # [START howto_operator_video_intelligence_detect_labels]
    detect_video_label = CloudVideoIntelligenceDetectVideoLabelsOperator(
        input_uri=INPUT_URI,
        output_uri=None,
        video_context=None,
        timeout=5,
        task_id="detect_video_label",
    )
    # [END howto_operator_video_intelligence_detect_labels]

    # [START howto_operator_video_intelligence_detect_labels_result]
    detect_video_label_result = BashOperator(
        bash_command="echo {{ task_instance.xcom_pull('detect_video_label')"
        "['annotationResults'][0]['shotLabelAnnotations'][0]['entity']}}",
        task_id="detect_video_label_result",
    )
    # [END howto_operator_video_intelligence_detect_labels_result]

    # [START howto_operator_video_intelligence_detect_explicit_content]
    detect_video_explicit_content = CloudVideoIntelligenceDetectVideoExplicitContentOperator(
        input_uri=INPUT_URI,
        output_uri=None,
        video_context=None,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="detect_video_explicit_content",
    )
    # [END howto_operator_video_intelligence_detect_explicit_content]

    # [START howto_operator_video_intelligence_detect_explicit_content_result]
    detect_video_explicit_content_result = BashOperator(
        bash_command="echo {{ task_instance.xcom_pull('detect_video_explicit_content')"
        "['annotationResults'][0]['explicitAnnotation']['frames'][0]}}",
        task_id="detect_video_explicit_content_result",
    )
    # [END howto_operator_video_intelligence_detect_explicit_content_result]

    # [START howto_operator_video_intelligence_detect_video_shots]
    detect_video_shots = CloudVideoIntelligenceDetectVideoShotsOperator(
        input_uri=INPUT_URI,
        output_uri=None,
        video_context=None,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="detect_video_shots",
    )
    # [END howto_operator_video_intelligence_detect_video_shots]

    # [START howto_operator_video_intelligence_detect_video_shots_result]
    detect_video_shots_result = BashOperator(
        bash_command="echo {{ task_instance.xcom_pull('detect_video_shots')"
        "['annotationResults'][0]['shotAnnotations'][0]}}",
        task_id="detect_video_shots_result",
    )
    # [END howto_operator_video_intelligence_detect_video_shots_result]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME_DST, trigger_rule=TriggerRule.ALL_DONE
    )

    chain(
        # TEST SETUP
        create_bucket,
        copy_single_file,
        # TEST BODY
        detect_video_label,
        detect_video_label_result,
        detect_video_explicit_content,
        detect_video_explicit_content_result,
        detect_video_shots,
        detect_video_shots_result,
        # TEST TEARDOWN
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.google.marketing_platform.hooks.search_ads import (
    GoogleSearchAdsHook,
    GoogleSearchAdsReportingHook,
)

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

GCP_CONN_ID = "google_cloud_default"
API_VERSION = "v0"
CUSTOMER_ID = "customer_id"
QUERY = "SELECT * FROM campaigns WHERE segments.date DURING LAST_30_DAYS"


class TestGoogleSearchAdsReportingHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = GoogleSearchAdsReportingHook(gcp_conn_id=GCP_CONN_ID)

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.get_credentials"
    )
    @mock.patch("airflow.providers.google.marketing_platform.hooks.search_ads.build")
    def test_gen_conn(self, mock_build, mock_get_credentials):
        result = self.hook.get_conn()
        mock_build.assert_called_once_with(
            "searchads360",
            API_VERSION,
            credentials=mock_get_credentials.return_value,
            cache_discovery=False,
        )
        assert mock_build.return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.customer_service"
    )
    @pytest.mark.parametrize(
        ("given_args", "expected_args_extras"),
        [
            ({"page_token": None}, {}),
            ({"page_token": "next_page_token"}, {"pageToken": "next_page_token"}),
            ({"summary_row_setting": "summary line content"}, {"summaryRowSetting": "summary line content"}),
            ({"page_size": 10, "validate_only": True}, {"pageSize": 10, "validateOnly": True}),
        ],
    )
    def test_search(self, customer_service_mock, given_args, expected_args_extras):
        return_value = {"results": [{"x": 1}]}
        (
            customer_service_mock.searchAds360.return_value.search.return_value.execute
        ).return_value = return_value

        result = self.hook.search(customer_id=CUSTOMER_ID, query=QUERY, **given_args)

        expected_args = {
            "customerId": CUSTOMER_ID,
            "body": {
                "query": QUERY,
                "pageSize": 10000,
                "returnTotalResultsCount": False,
                "validateOnly": False,
                **expected_args_extras,
            },
        }
        customer_service_mock.searchAds360.return_value.search.assert_called_once_with(**expected_args)

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.customer_service"
    )
    def test_get_custom_column(self, customer_service_mock):
        custom_column_id = "custom_column_id"
        return_value = {"resourceName": 1}
        (
            customer_service_mock.customColumns.return_value.get.return_value.execute
        ).return_value = return_value

        result = self.hook.get_custom_column(customer_id=CUSTOMER_ID, custom_column_id=custom_column_id)

        customer_service_mock.customColumns.return_value.get.assert_called_once_with(
            resourceName=f"customers/{CUSTOMER_ID}/customColumns/{custom_column_id}"
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.customer_service"
    )
    def test_list_custom_columns(self, customer_service_mock):
        return_value = {
            "results": [
                {"resourceName": f"customers/{CUSTOMER_ID}/customColumns/col1"},
                {"resourceName": f"customers/{CUSTOMER_ID}/customColumns/col2"},
            ]
        }
        (
            customer_service_mock.customColumns.return_value.list.return_value.execute
        ).return_value = return_value

        result = self.hook.list_custom_columns(customer_id=CUSTOMER_ID)

        customer_service_mock.customColumns.return_value.list.assert_called_once_with(customerId=CUSTOMER_ID)

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.fields_service"
    )
    def test_get_field(self, fields_service_mock):
        field_name = "field_name"
        return_value = {
            "name": "Field 1",
            "resourceName": f"customers/{CUSTOMER_ID}/searchAds360Fields/field1",
        }
        fields_service_mock.get.return_value.execute.return_value = return_value

        result = self.hook.get_field(field_name=field_name)

        fields_service_mock.get.assert_called_once_with(resourceName=f"searchAds360Fields/{field_name}")

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsReportingHook.fields_service"
    )
    @pytest.mark.parametrize(
        ("given_args", "expected_args_extras"),
        [
            ({"page_token": None}, {}),
            ({"page_token": "next_page_token"}, {"pageToken": "next_page_token"}),
            ({"page_size": 10}, {"pageSize": 10}),
        ],
    )
    def test_search_fields(self, fields_service_mock, given_args, expected_args_extras):
        query = "SELECT field1, field2 FROM campaigns;"
        return_value = {
            "results": [
                {"name": "Field 1", "resourceName": f"customers/{CUSTOMER_ID}/searchAds360Fields/field1"},
                {"name": "Field 2", "resourceName": f"customers/{CUSTOMER_ID}/searchAds360Fields/field2"},
            ]
        }
        fields_service_mock.search.return_value.execute.return_value = return_value

        result = self.hook.search_fields(query=query, **given_args)

        expected_args = {"query": query, "pageSize": 10000, **expected_args_extras}
        fields_service_mock.search.assert_called_once_with(body=expected_args)

        assert return_value == result


class TestSearchAdsHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.marketing_platform.hooks.search_ads.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = GoogleSearchAdsHook(gcp_conn_id=GCP_CONN_ID)

    @mock.patch("airflow.providers.google.marketing_platform.hooks.search_ads.GoogleSearchAdsHook._authorize")
    @mock.patch("airflow.providers.google.marketing_platform.hooks.search_ads.build")
    def test_gen_conn(self, mock_build, mock_authorize):
        result = self.hook.get_conn()
        mock_build.assert_called_once_with(
            "doubleclicksearch",
            "v2",
            http=mock_authorize.return_value,
            cache_discovery=False,
        )
        assert mock_build.return_value == result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Storage GCSSynchronizeBucketsOperator and
GCSToGCSOperator operators.
"""

from __future__ import annotations

import os
import shutil
from datetime import datetime
from pathlib import Path

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSListObjectsOperator,
    GCSSynchronizeBucketsOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator
from airflow.providers.google.cloud.transfers.local_to_gcs import LocalFilesystemToGCSOperator
from airflow.providers.standard.operators.bash import BashOperator
from airflow.providers.standard.operators.python import PythonOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from system.openlineage.operator import OpenLineageTestOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "gcs_to_gcs"

BUCKET_NAME_SRC = f"bucket_{DAG_ID}_{ENV_ID}"
BUCKET_NAME_DST = f"bucket_dst_{DAG_ID}_{ENV_ID}"
RANDOM_FILE_NAME = OBJECT_1 = OBJECT_2 = "random.bin"
HOME = "/home/airflow/gcs"
PREFIX = f"{HOME}/data/{DAG_ID}_{ENV_ID}/"


def _assert_copied_files_exist(ti):
    objects = ti.xcom_pull("list_objects")

    assert PREFIX + OBJECT_1 in objects
    assert f"{PREFIX}subdir/{OBJECT_1}" in objects
    assert f"backup/{OBJECT_1}" in objects
    assert f"backup_{OBJECT_1}" in objects


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["gcs", "example"],
) as dag:

    @task
    def create_workdir() -> str:
        """
        Task creates working directory. The logic behind this task is a workaround that provides sustainable
        execution in Composer environment: local files can be safely shared among tasks if they are located
        within '/home/airflow/gcs/data/' folder which is mounted to GCS bucket under the hood
        (https://cloud.google.com/composer/docs/composer-2/cloud-storage).
        Otherwise, worker nodes don't share local path and thus files created by one task aren't guaranteed
        to be accessible be others.
        """
        workdir = PREFIX if os.path.exists(HOME) else HOME
        os.makedirs(PREFIX)
        return workdir

    create_workdir_task = create_workdir()

    generate_random_file = BashOperator(
        task_id="generate_random_file",
        bash_command=f"cat /dev/urandom | head -c $((1 * 1024 * 1024)) > {PREFIX + RANDOM_FILE_NAME}",
    )

    create_bucket_src = GCSCreateBucketOperator(
        task_id="create_bucket_src",
        bucket_name=BUCKET_NAME_SRC,
        project_id=PROJECT_ID,
    )

    create_bucket_dst = GCSCreateBucketOperator(
        task_id="create_bucket_dst",
        bucket_name=BUCKET_NAME_DST,
        project_id=PROJECT_ID,
    )

    upload_file_src = LocalFilesystemToGCSOperator(
        task_id="upload_file_src",
        src=PREFIX + RANDOM_FILE_NAME,
        dst=PREFIX + RANDOM_FILE_NAME,
        bucket=BUCKET_NAME_SRC,
    )

    upload_file_src_sub = LocalFilesystemToGCSOperator(
        task_id="upload_file_src_sub",
        src=PREFIX + RANDOM_FILE_NAME,
        dst=f"{PREFIX}subdir/{RANDOM_FILE_NAME}",
        bucket=BUCKET_NAME_SRC,
    )

    upload_file_dst = LocalFilesystemToGCSOperator(
        task_id="upload_file_dst",
        src=PREFIX + RANDOM_FILE_NAME,
        dst=PREFIX + RANDOM_FILE_NAME,
        bucket=BUCKET_NAME_DST,
    )

    upload_file_dst_sub = LocalFilesystemToGCSOperator(
        task_id="upload_file_dst_sub",
        src=PREFIX + RANDOM_FILE_NAME,
        dst=f"{PREFIX}subdir/{RANDOM_FILE_NAME}",
        bucket=BUCKET_NAME_DST,
    )

    # [START howto_synch_bucket]
    sync_bucket = GCSSynchronizeBucketsOperator(
        task_id="sync_bucket", source_bucket=BUCKET_NAME_SRC, destination_bucket=BUCKET_NAME_DST
    )
    # [END howto_synch_bucket]

    # [START howto_synch_full_bucket]
    sync_full_bucket = GCSSynchronizeBucketsOperator(
        task_id="sync_full_bucket",
        source_bucket=BUCKET_NAME_SRC,
        destination_bucket=BUCKET_NAME_DST,
        delete_extra_files=True,
        allow_overwrite=True,
    )
    # [END howto_synch_full_bucket]

    # [START howto_synch_to_subdir]
    sync_to_subdirectory = GCSSynchronizeBucketsOperator(
        task_id="sync_to_subdirectory",
        source_bucket=BUCKET_NAME_SRC,
        destination_bucket=BUCKET_NAME_DST,
        destination_object="subdir/",
    )
    # [END howto_synch_to_subdir]

    # [START howto_sync_from_subdir]
    sync_from_subdirectory = GCSSynchronizeBucketsOperator(
        task_id="sync_from_subdirectory",
        source_bucket=BUCKET_NAME_SRC,
        source_object="subdir/",
        destination_bucket=BUCKET_NAME_DST,
    )
    # [END howto_sync_from_subdir]

    # [START howto_operator_gcs_to_gcs_single_file]
    copy_single_file = GCSToGCSOperator(
        task_id="copy_single_gcs_file",
        source_bucket=BUCKET_NAME_SRC,
        source_object=PREFIX + OBJECT_1,
        destination_bucket=BUCKET_NAME_DST,  # If not supplied the source_bucket value will be used
        destination_object="backup_" + OBJECT_1,  # If not supplied the source_object value will be used
        exact_match=True,
    )
    # [END howto_operator_gcs_to_gcs_single_file]

    # [START howto_operator_gcs_to_gcs_without_wildcard]
    copy_files = GCSToGCSOperator(
        task_id="copy_files",
        source_bucket=BUCKET_NAME_SRC,
        source_object=PREFIX + "subdir/",
        destination_bucket=BUCKET_NAME_DST,
        destination_object="backup/",
    )
    # [END howto_operator_gcs_to_gcs_without_wildcard]

    # [START howto_operator_gcs_to_gcs_match_glob]
    copy_files_with_match_glob = GCSToGCSOperator(
        task_id="copy_files_with_match_glob",
        source_bucket=BUCKET_NAME_SRC,
        source_object="data/",
        destination_bucket=BUCKET_NAME_DST,
        destination_object="backup/",
        match_glob="**/*.txt",
    )
    # [END howto_operator_gcs_to_gcs_match_glob]

    # [START howto_operator_gcs_to_gcs_list]
    copy_files_with_list = GCSToGCSOperator(
        task_id="copy_files_with_list",
        source_bucket=BUCKET_NAME_SRC,
        source_objects=[
            PREFIX + OBJECT_1,
            PREFIX + OBJECT_2,
        ],  # Instead of files each element could be a wildcard expression
        destination_bucket=BUCKET_NAME_DST,
        destination_object="backup/",
    )
    # [END howto_operator_gcs_to_gcs_list]

    # [START howto_operator_gcs_to_gcs_single_file_move]
    move_single_file = GCSToGCSOperator(
        task_id="move_single_file",
        source_bucket=BUCKET_NAME_SRC,
        source_object=PREFIX + OBJECT_1,
        destination_bucket=BUCKET_NAME_DST,
        destination_object="backup_" + OBJECT_1,
        exact_match=True,
        move_object=True,
    )
    # [END howto_operator_gcs_to_gcs_single_file_move]

    # [START howto_operator_gcs_to_gcs_list_move]
    move_files_with_list = GCSToGCSOperator(
        task_id="move_files_with_list",
        source_bucket=BUCKET_NAME_SRC,
        source_objects=[PREFIX + OBJECT_1, PREFIX + OBJECT_2],
        destination_bucket=BUCKET_NAME_DST,
        destination_object="backup/",
    )
    # [END howto_operator_gcs_to_gcs_list_move]

    list_objects = GCSListObjectsOperator(task_id="list_objects", bucket=BUCKET_NAME_DST)
    assert_copied_files_exist = PythonOperator(
        task_id="assert_copied_files_exist",
        python_callable=_assert_copied_files_exist,
    )

    delete_bucket_src = GCSDeleteBucketOperator(
        task_id="delete_bucket_src", bucket_name=BUCKET_NAME_SRC, trigger_rule=TriggerRule.ALL_DONE
    )
    delete_bucket_dst = GCSDeleteBucketOperator(
        task_id="delete_bucket_dst", bucket_name=BUCKET_NAME_DST, trigger_rule=TriggerRule.ALL_DONE
    )

    check_openlineage_events = OpenLineageTestOperator(
        task_id="check_openlineage_events",
        file_path=str(Path(__file__).parent / "resources" / "openlineage" / "gcs_to_gcs.json"),
    )

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_work_dir(create_workdir_result: str) -> None:
        shutil.rmtree(create_workdir_result)

    chain(
        # TEST SETUP
        create_workdir_task,
        generate_random_file,
        [create_bucket_src, create_bucket_dst],
        [upload_file_src, upload_file_src_sub],
        [upload_file_dst, upload_file_dst_sub],
        # TEST BODY
        sync_bucket,
        sync_full_bucket,
        sync_to_subdirectory,
        sync_from_subdirectory,
        copy_single_file,
        copy_files,
        copy_files_with_match_glob,
        copy_files_with_list,
        move_single_file,
        move_files_with_list,
        list_objects,
        assert_copied_files_exist,
        # TEST TEARDOWN
        [
            delete_bucket_src,
            delete_bucket_dst,
            delete_work_dir(create_workdir_task),
            check_openlineage_events,
        ],
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import inspect
import re
from collections.abc import Callable
from datetime import date, datetime
from typing import Any

from deprecated import deprecated as standard_deprecated
from deprecated.classic import ClassicAdapter


class AirflowDeprecationAdapter(ClassicAdapter):
    """
    Build a detailed deprecation message based on the wrapped object type and other provided details.

    :param planned_removal_date: The date after which the deprecated object should be removed.
        The recommended date is six months ahead from today. The expected date format is `Month DD, YYYY`,
        for example: `August 22, 2024`.
        This parameter is required if the `planned_removal_release` parameter is not set.
    :param planned_removal_release: The package name and the version in which the deprecated object is
        expected to be removed. The expected format is `<package_name>==<package_version>`, for example
        `apache-airflow==2.10.0` or `apache-airflow-providers-google==10.22.0`.
        This parameter is required if the `planned_removal_date` parameter is not set.
    :param use_instead: Optional. Replacement of the deprecated object.
    :param reason: Optional. The detailed reason for the deprecated object.
    :param instructions: Optional. The detailed instructions for migrating from the deprecated object.
    :param category: Optional. The warning category to be used for the deprecation warning.
    """

    def __init__(
        self,
        planned_removal_date: str | None = None,
        planned_removal_release: str | None = None,
        use_instead: str | None = None,
        reason: str | None = None,
        instructions: str | None = None,
        category: type[DeprecationWarning] = DeprecationWarning,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.planned_removal_date: date | None = self._validate_date(planned_removal_date)
        self.planned_removal_release: str | None = self._validate_removal_release(planned_removal_release)
        self.use_instead: str | None = use_instead
        self.reason: str = reason or ""
        self.instructions: str | None = instructions
        self.category: type[DeprecationWarning] = category
        self._validate_fields()

    def get_deprecated_msg(self, wrapped: Callable, instance: Any):
        """
        Generate a deprecation message for wrapped callable.

        :param wrapped: Deprecated entity.
        :param instance: The instance to which the callable belongs. (not used)
        :return: A formatted deprecation message with all the details.
        """
        entity_type = self.entity_type(entity=wrapped)
        entity_path = self.entity_path(entity=wrapped)
        sunset = self.sunset_message()
        replacement = self.replacement_message()
        msg = f"The {entity_type} `{entity_path}` is deprecated and will be removed {sunset}. {replacement}"
        if self.reason:
            msg += f" The reason is: {self.reason}"
        if self.instructions:
            msg += f" Instructions: {self.instructions}"
        return msg

    @staticmethod
    def _validate_date(value: str | None) -> date | None:
        if value:
            try:
                return datetime.strptime(value, "%B %d, %Y").date()
            except ValueError as ex:
                error_message = (
                    f"Invalid date '{value}'. "
                    f"The expected format is 'Month DD, YYYY', for example 'August 22, 2024'."
                )
                raise ValueError(error_message) from ex
        return None

    @staticmethod
    def _validate_removal_release(value: str | None) -> str | None:
        if value:
            pattern = r"^apache-airflow(-providers-[a-zA-Z-]+)?==\d+\.\d+\.\d+.*$"
            if not bool(re.match(pattern, value)):
                raise ValueError(
                    f"`{value}` must follow the format 'apache-airflow(-providers-<name>)==<X.Y.Z>'."
                )
        return value

    def _validate_fields(self):
        msg = "Only one of two parameters must be set: `planned_removal_date` or 'planned_removal_release'."
        if self.planned_removal_release and self.planned_removal_date:
            raise ValueError(f"{msg} You specified both.")

    @staticmethod
    def entity_type(entity: Callable) -> str:
        return "class" if inspect.isclass(entity) else "function (or method)"

    @staticmethod
    def entity_path(entity: Callable) -> str:
        module_name = getattr(entity, "__module__", "")
        qualified_name = getattr(entity, "__qualname__", "")
        full_path = f"{module_name}.{qualified_name}".strip(".")

        if module_name and full_path:
            return full_path
        return str(entity)

    def sunset_message(self) -> str:
        if self.planned_removal_date:
            return f"after {self.planned_removal_date.strftime('%B %d, %Y')}"
        if self.planned_removal_release:
            return f"since version {self.planned_removal_release}"
        return "in the future"

    def replacement_message(self):
        if self.use_instead:
            replacements = ", ".join(f"`{replacement}`" for replacement in self.use_instead.split(", "))
            return f"Please use {replacements} instead."
        return "There is no replacement."


def deprecated(
    *args,
    planned_removal_date: str | None = None,
    planned_removal_release: str | None = None,
    use_instead: str | None = None,
    reason: str | None = None,
    instructions: str | None = None,
    adapter_cls: type[AirflowDeprecationAdapter] = AirflowDeprecationAdapter,
    **kwargs,
):
    """
    Mark a class, method or a function deprecated.

    :param planned_removal_date: The date after which the deprecated object should be removed.
        The recommended date is six months ahead from today. The expected date format is `Month DD, YYYY`,
        for example: `August 22, 2024`.
        This parameter is required if the `planned_removal_release` parameter is not set.
    :param planned_removal_release: The package name and the version in which the deprecated object is
        expected to be removed. The expected format is `<package_name>==<package_version>`, for example
        `apache-airflow==2.10.0` or `apache-airflow-providers-google==10.22.0`.
        This parameter is required if the `planned_removal_date` parameter is not set.
    :param use_instead: Optional. Replacement of the deprecated object.
    :param reason: Optional. The detailed reason for the deprecated object.
    :param instructions: Optional. The detailed instructions for migrating from the deprecated object.
    :param adapter_cls: Optional. Adapter class that is used to get the deprecation message
        This should be a subclass of `AirflowDeprecationAdapter`.
    """
    _kwargs = {
        **kwargs,
        "planned_removal_date": planned_removal_date,
        "planned_removal_release": planned_removal_release,
        "use_instead": use_instead,
        "reason": reason,
        "instructions": instructions,
        "adapter_cls": adapter_cls,
    }
    return standard_deprecated(*args, **_kwargs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowOptionalProviderFeatureException

try:
    from airflow.providers.google.leveldb.hooks.leveldb import LevelDBHook, LevelDBHookException
except AirflowOptionalProviderFeatureException:
    pytest.skip("LevelDB not available", allow_module_level=True)


class TestLevelDBHook:
    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_get_conn_db_is_not_none(self, tmp_path):
        """Test get_conn method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        assert hook.db is not None, "Check existence of DB object in connection creation"
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_run(self, tmp_path):
        """Test run method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        assert hook.run("get", b"test_key0") is None, "Initially, this key in LevelDB is empty"
        hook.run("put", b"test_key0", b"test_value0")
        assert hook.run("get", b"test_key0") == b"test_value0", (
            "Connection to LevelDB with PUT and GET works."
        )
        hook.run("delete", b"test_key0")
        assert hook.run("get", b"test_key0") is None, "Connection to LevelDB with DELETE works."
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_get(self, tmp_path):
        """Test get method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        db = hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        db.put(b"test_key", b"test_value")
        assert hook.get(b"test_key") == b"test_value"
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_put(self, tmp_path):
        """Test put method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        db = hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        hook.put(b"test_key2", b"test_value2")
        assert db.get(b"test_key2") == b"test_value2"
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_delete(self, tmp_path):
        """Test delete method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        db = hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        db.put(b"test_key3", b"test_value3")
        hook.delete(b"test_key3")
        assert db.get(b"test_key3") is None
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_write_batch(self, tmp_path):
        """Test write batch method of hook"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        db = hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        keys = [b"key", b"another-key"]
        values = [b"value", b"another-value"]
        hook.write_batch(keys, values)
        assert db.get(b"key") == b"value"
        assert db.get(b"another-key") == b"another-value"
        hook.close_conn()

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_exception(self, tmp_path):
        """Test raising exception of hook in run method if we have unknown command in input"""
        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        hook.get_conn(name=tmp_path.as_posix(), create_if_missing=True)
        with pytest.raises(LevelDBHookException):
            hook.run(command="other_command", key=b"key", value=b"value")

    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    def test_comparator(self, tmp_path):
        """Test comparator"""

        def comparator(a, b):
            a = a.lower()
            b = b.lower()

            if a < b:
                # a sorts before b
                return -1

            if a > b:
                # a sorts after b
                return 1

            # a and b are equal
            return 0

        hook = LevelDBHook(leveldb_conn_id="leveldb_default")
        hook.get_conn(
            name=tmp_path.as_posix(),
            create_if_missing=True,
            comparator=comparator,
            comparator_name=b"CaseInsensitiveComparator",
        )
        assert hook.db is not None, "Check existence of DB object(with comparator) in connection creation"
        hook.close_conn()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that:
* creates a copy of existing Instance Template
* updates existing template in Instance Group Manager

"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.compute import (
    ComputeEngineCopyInstanceTemplateOperator,
    ComputeEngineDeleteInstanceGroupManagerOperator,
    ComputeEngineDeleteInstanceTemplateOperator,
    ComputeEngineInsertInstanceGroupManagerOperator,
    ComputeEngineInsertInstanceTemplateOperator,
    ComputeEngineInstanceGroupUpdateManagerTemplateOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

LOCATION = "europe-west1-b"
REGION = "europe-west1"
SHORT_MACHINE_TYPE_NAME = "n1-standard-1"
DAG_ID = "cloud_compute_igm"

# [START howto_operator_compute_template_copy_args]
TEMPLATE_NAME = "instance-template-igm-test"
NEW_TEMPLATE_NAME = "instance-template-test-new"

INSTANCE_TEMPLATE_BODY = {
    "name": TEMPLATE_NAME,
    "properties": {
        "machine_type": SHORT_MACHINE_TYPE_NAME,
        "disks": [
            {
                "auto_delete": True,
                "boot": True,
                "device_name": TEMPLATE_NAME,
                "initialize_params": {
                    "disk_size_gb": "10",
                    "disk_type": "pd-balanced",
                    "source_image": "projects/debian-cloud/global/images/debian-12-bookworm-v20240611",
                },
            }
        ],
        "network_interfaces": [{"network": "global/networks/default"}],
    },
}

NEW_DESCRIPTION = "Test new description"
INSTANCE_TEMPLATE_BODY_UPDATE = {
    "name": NEW_TEMPLATE_NAME,
    "description": NEW_DESCRIPTION,
    "properties": {"machine_type": "n1-standard-2"},
}
# [END howto_operator_compute_template_copy_args]

# [START howto_operator_compute_igm_update_template_args]
INSTANCE_GROUP_MANAGER_NAME = "instance-group-test"
INSTANCE_GROUP_MANAGER_BODY = {
    "name": INSTANCE_GROUP_MANAGER_NAME,
    "base_instance_name": INSTANCE_GROUP_MANAGER_NAME,
    "instance_template": f"global/instanceTemplates/{TEMPLATE_NAME}",
    "target_size": 1,
}

SOURCE_TEMPLATE_URL = (
    f"https://www.googleapis.com/compute/beta/projects/{PROJECT_ID}/global/instanceTemplates/{TEMPLATE_NAME}"
)


DESTINATION_TEMPLATE_URL = (
    f"https://www.googleapis.com/compute/beta/projects/{PROJECT_ID}/"
    f"global/instanceTemplates/{NEW_TEMPLATE_NAME}"
)

UPDATE_POLICY = {
    "type": "OPPORTUNISTIC",
    "minimalAction": "RESTART",
    "maxSurge": {"fixed": 1},
    "minReadySec": 1800,
}
# [END howto_operator_compute_igm_update_template_args]


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "compute-igm"],
) as dag:
    # [START howto_operator_gce_igm_insert_template]
    gce_instance_template_insert = ComputeEngineInsertInstanceTemplateOperator(
        task_id="gcp_compute_create_template_task",
        project_id=PROJECT_ID,
        body=INSTANCE_TEMPLATE_BODY,
    )
    # [END howto_operator_gce_igm_insert_template]

    # Added to check for idempotence
    # [START howto_operator_gce_igm_insert_template_no_project_id]
    gce_instance_template_insert2 = ComputeEngineInsertInstanceTemplateOperator(
        task_id="gcp_compute_create_template_task_2",
        body=INSTANCE_TEMPLATE_BODY,
    )
    # [END howto_operator_gce_igm_insert_template_no_project_id]

    # [START howto_operator_gce_igm_copy_template]
    gce_instance_template_copy = ComputeEngineCopyInstanceTemplateOperator(
        task_id="gcp_compute_igm_copy_template_task",
        project_id=PROJECT_ID,
        resource_id=TEMPLATE_NAME,
        body_patch=INSTANCE_TEMPLATE_BODY_UPDATE,
    )
    # [END howto_operator_gce_igm_copy_template]

    # Added to check for idempotence
    # [START howto_operator_gce_igm_copy_template_no_project_id]
    gce_instance_template_copy2 = ComputeEngineCopyInstanceTemplateOperator(
        task_id="gcp_compute_igm_copy_template_task_2",
        resource_id=TEMPLATE_NAME,
        body_patch=INSTANCE_TEMPLATE_BODY_UPDATE,
    )
    # [END howto_operator_gce_igm_copy_template_no_project_id]

    # [START howto_operator_gce_insert_igm]
    gce_igm_insert = ComputeEngineInsertInstanceGroupManagerOperator(
        task_id="gcp_compute_create_group_task",
        zone=LOCATION,
        body=INSTANCE_GROUP_MANAGER_BODY,
        project_id=PROJECT_ID,
    )
    # [END howto_operator_gce_insert_igm]

    # Added to check for idempotence
    # [START howto_operator_gce_insert_igm_no_project_id]
    gce_igm_insert2 = ComputeEngineInsertInstanceGroupManagerOperator(
        task_id="gcp_compute_create_group_task_2",
        zone=LOCATION,
        body=INSTANCE_GROUP_MANAGER_BODY,
    )
    # [END howto_operator_gce_insert_igm_no_project_id]

    # [START howto_operator_gce_igm_update_template]
    gce_instance_group_manager_update_template = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
        task_id="gcp_compute_igm_group_manager_update_template",
        project_id=PROJECT_ID,
        resource_id=INSTANCE_GROUP_MANAGER_NAME,
        zone=LOCATION,
        source_template=SOURCE_TEMPLATE_URL,
        destination_template=DESTINATION_TEMPLATE_URL,
        update_policy=UPDATE_POLICY,
    )
    # [END howto_operator_gce_igm_update_template]

    # Added to check for idempotence (and without UPDATE_POLICY)
    # [START howto_operator_gce_igm_update_template_no_project_id]
    gce_instance_group_manager_update_template2 = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
        task_id="gcp_compute_igm_group_manager_update_template_2",
        resource_id=INSTANCE_GROUP_MANAGER_NAME,
        zone=LOCATION,
        source_template=SOURCE_TEMPLATE_URL,
        destination_template=DESTINATION_TEMPLATE_URL,
    )
    # [END howto_operator_gce_igm_update_template_no_project_id]

    # [START howto_operator_gce_delete_old_template_no_project_id]
    gce_instance_template_old_delete = ComputeEngineDeleteInstanceTemplateOperator(
        task_id="gcp_compute_delete_old_template_task",
        resource_id=TEMPLATE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_gce_delete_old_template_no_project_id]

    # [START howto_operator_gce_delete_new_template_no_project_id]
    gce_instance_template_new_delete = ComputeEngineDeleteInstanceTemplateOperator(
        task_id="gcp_compute_delete_new_template_task",
        resource_id=NEW_TEMPLATE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_gce_delete_new_template_no_project_id]

    # [START howto_operator_gce_delete_igm_no_project_id]
    gce_igm_delete = ComputeEngineDeleteInstanceGroupManagerOperator(
        task_id="gcp_compute_delete_group_task",
        resource_id=INSTANCE_GROUP_MANAGER_NAME,
        zone=LOCATION,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_gce_delete_igm_no_project_id]

    (
        # TEST SETUP
        gce_instance_template_insert
        >> gce_instance_template_insert2
        >> gce_instance_template_copy
        >> gce_instance_template_copy2
        # TEST BODY
        >> gce_igm_insert
        >> gce_igm_insert2
        >> gce_instance_group_manager_update_template
        >> gce_instance_group_manager_update_template2
        # TEST TEARDOWN
        >> gce_igm_delete
        >> gce_instance_template_old_delete
        >> gce_instance_template_new_delete
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Cloud AlloyDB links."""

from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

ALLOY_DB_BASE_LINK = "/alloydb"
ALLOY_DB_CLUSTER_LINK = (
    ALLOY_DB_BASE_LINK + "/locations/{location_id}/clusters/{cluster_id}?project={project_id}"
)
ALLOY_DB_USERS_LINK = (
    ALLOY_DB_BASE_LINK + "/locations/{location_id}/clusters/{cluster_id}/users?project={project_id}"
)
ALLOY_DB_BACKUPS_LINK = ALLOY_DB_BASE_LINK + "/backups?project={project_id}"


class AlloyDBClusterLink(BaseGoogleLink):
    """Helper class for constructing AlloyDB cluster Link."""

    name = "AlloyDB Cluster"
    key = "alloy_db_cluster"
    format_str = ALLOY_DB_CLUSTER_LINK


class AlloyDBUsersLink(BaseGoogleLink):
    """Helper class for constructing AlloyDB users Link."""

    name = "AlloyDB Users"
    key = "alloy_db_users"
    format_str = ALLOY_DB_USERS_LINK


class AlloyDBBackupsLink(BaseGoogleLink):
    """Helper class for constructing AlloyDB backups Link."""

    name = "AlloyDB Backups"
    key = "alloy_db_backups"
    format_str = ALLOY_DB_BACKUPS_LINK



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
import os
from tempfile import TemporaryDirectory

import pytest

from airflow.providers.google.cloud.hooks.kms import CloudKMSHook

from tests_common.test_utils.gcp_system_helpers import GoogleSystemTest, provide_gcp_context
from unit.google.cloud.utils.gcp_authenticator import GCP_KMS_KEY

# To prevent resource name collisions, key ring and key resources CANNOT be deleted, so
# to avoid cluttering the project, we only create the key once during project initialization.
# See: https://cloud.google.com/kms/docs/faq#cannot_delete
GCP_KMS_KEYRING_NAME = os.environ.get("GCP_KMS_KEYRING_NAME", "test-airflow-system-tests-keyring")
GCP_KMS_KEY_NAME = os.environ.get("GCP_KMS_KEY_NAME", "test-airflow-system-tests-key")


@pytest.mark.credential_file(GCP_KMS_KEY)
class TestKmsHookSystem(GoogleSystemTest):
    @provide_gcp_context(GCP_KMS_KEY)
    def test_encrypt(self):
        with TemporaryDirectory() as tmp_dir:
            kms_hook = CloudKMSHook()
            content = kms_hook.encrypt(
                key_name=(
                    f"projects/{kms_hook.project_id}/locations/global/keyRings/"
                    f"{GCP_KMS_KEYRING_NAME}/cryptoKeys/{GCP_KMS_KEY_NAME}"
                ),
                plaintext=b"TEST-SECRET",
            )
            with open(f"{tmp_dir}/mysecret.txt.encrypted", "wb") as encrypted_file:
                encrypted_file.write(base64.b64decode(content))
            self.execute_cmd(
                [
                    "gcloud",
                    "kms",
                    "decrypt",
                    "--location",
                    "global",
                    "--keyring",
                    GCP_KMS_KEYRING_NAME,
                    "--key",
                    GCP_KMS_KEY_NAME,
                    "--ciphertext-file",
                    f"{tmp_dir}/mysecret.txt.encrypted",
                    "--plaintext-file",
                    f"{tmp_dir}/mysecret.txt",
                ]
            )
            with open(f"{tmp_dir}/mysecret.txt", "rb") as secret_file:
                secret = secret_file.read()
            assert secret == b"TEST-SECRET"

    @provide_gcp_context(GCP_KMS_KEY)
    def test_decrypt(self):
        with TemporaryDirectory() as tmp_dir:
            with open(f"{tmp_dir}/mysecret.txt", "w") as secret_file:
                secret_file.write("TEST-SECRET")
            self.execute_cmd(
                [
                    "gcloud",
                    "kms",
                    "encrypt",
                    "--location",
                    "global",
                    "--keyring",
                    GCP_KMS_KEYRING_NAME,
                    "--key",
                    GCP_KMS_KEY_NAME,
                    "--plaintext-file",
                    f"{tmp_dir}/mysecret.txt",
                    "--ciphertext-file",
                    f"{tmp_dir}/mysecret.txt.encrypted",
                ]
            )
            with open(f"{tmp_dir}/mysecret.txt.encrypted", "rb") as encrypted_file:
                encrypted_secret = base64.b64encode(encrypted_file.read()).decode()

            kms_hook = CloudKMSHook()
            content = kms_hook.decrypt(
                key_name=(
                    f"projects/{kms_hook.project_id}/locations/global/keyRings/"
                    f"{GCP_KMS_KEYRING_NAME}/cryptoKeys/{GCP_KMS_KEY_NAME}"
                ),
                ciphertext=encrypted_secret,
            )
            assert content == b"TEST-SECRET"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from airflow.providers.google.suite.hooks.sheets import GSheetsHook
from airflow.providers.google.version_compat import BaseOperator


class GoogleSheetsCreateSpreadsheetOperator(BaseOperator):
    """
    Creates a new spreadsheet.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleSheetsCreateSpreadsheetOperator`

    :param spreadsheet: an instance of Spreadsheet
        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#Spreadsheet
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param api_endpoint: Optional. Custom API endpoint, e.g: private.googleapis.com.
        This can be used to target private VPC or restricted access endpoints.
    """

    template_fields: Sequence[str] = (
        "spreadsheet",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        spreadsheet: dict[str, Any],
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        api_endpoint: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.spreadsheet = spreadsheet
        self.impersonation_chain = impersonation_chain
        self.api_endpoint = api_endpoint

    def execute(self, context: Any) -> dict[str, Any]:
        hook = GSheetsHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            api_endpoint=self.api_endpoint,
        )
        spreadsheet = hook.create_spreadsheet(spreadsheet=self.spreadsheet)
        context["task_instance"].xcom_push(key="spreadsheet_id", value=spreadsheet["spreadsheetId"])
        context["task_instance"].xcom_push(key="spreadsheet_url", value=spreadsheet["spreadsheetUrl"])
        return spreadsheet



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains operator for copying data from Cassandra to Google Cloud Storage in JSON format."""

from __future__ import annotations

import json
from base64 import b64encode
from collections.abc import Iterable, Sequence
from datetime import datetime
from decimal import Decimal
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, NewType
from uuid import UUID

from cassandra.util import Date, OrderedMapSerializedKey, SortedSet, Time

from airflow.exceptions import AirflowException
from airflow.providers.apache.cassandra.hooks.cassandra import CassandraHook
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context

NotSetType = NewType("NotSetType", object)
NOT_SET = NotSetType(object())


class CassandraToGCSOperator(BaseOperator):
    """
    Copy data from Cassandra to Google Cloud Storage in JSON format.

    Note: Arrays of arrays are not supported.

    :param cql: The CQL to execute on the Cassandra table.
    :param bucket: The bucket to upload to.
    :param filename: The filename to use as the object name when uploading
        to Google Cloud Storage. A {} should be specified in the filename
        to allow the operator to inject file numbers in cases where the
        file is split due to size.
    :param schema_filename: If set, the filename to use as the object name
        when uploading a .json file containing the BigQuery schema fields
        for the table that was dumped from MySQL.
    :param approx_max_file_size_bytes: This operator supports the ability
        to split large table dumps into multiple files (see notes in the
        filename param docs above). This param allows developers to specify the
        file size of the splits. Check https://cloud.google.com/storage/quotas
        to see the maximum allowed file size for a single object.
    :param cassandra_conn_id: Reference to a specific Cassandra hook.
    :param gzip: Option to compress file for upload
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param query_timeout: (Optional) The amount of time, in seconds, used to execute the Cassandra query.
        If not set, the timeout value will be set in Session.execute() by Cassandra driver.
        If set to None, there is no timeout.
    :param encode_uuid: (Optional) Option to encode UUID or not when upload from Cassandra to GCS.
        Default is to encode UUID.
    """

    template_fields: Sequence[str] = (
        "cql",
        "bucket",
        "filename",
        "schema_filename",
        "impersonation_chain",
    )
    template_ext: Sequence[str] = (".cql",)
    ui_color = "#a0e08c"

    def __init__(
        self,
        *,
        cql: str,
        bucket: str,
        filename: str,
        schema_filename: str | None = None,
        approx_max_file_size_bytes: int = 1900000000,
        gzip: bool = False,
        cassandra_conn_id: str = "cassandra_default",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        query_timeout: float | None | NotSetType = NOT_SET,
        encode_uuid: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.cql = cql
        self.bucket = bucket
        self.filename = filename
        self.schema_filename = schema_filename
        self.approx_max_file_size_bytes = approx_max_file_size_bytes
        self.cassandra_conn_id = cassandra_conn_id
        self.gcp_conn_id = gcp_conn_id
        self.gzip = gzip
        self.impersonation_chain = impersonation_chain
        self.query_timeout = query_timeout
        self.encode_uuid = encode_uuid

    # Default Cassandra to BigQuery type mapping
    CQL_TYPE_MAP = {
        "BytesType": "STRING",
        "DecimalType": "FLOAT",
        "UUIDType": "STRING",
        "BooleanType": "BOOL",
        "ByteType": "INTEGER",
        "AsciiType": "STRING",
        "FloatType": "FLOAT",
        "DoubleType": "FLOAT",
        "LongType": "INTEGER",
        "Int32Type": "INTEGER",
        "IntegerType": "INTEGER",
        "InetAddressType": "STRING",
        "CounterColumnType": "INTEGER",
        "DateType": "TIMESTAMP",
        "SimpleDateType": "DATE",
        "TimestampType": "TIMESTAMP",
        "TimeUUIDType": "STRING",
        "ShortType": "INTEGER",
        "TimeType": "TIME",
        "DurationType": "INTEGER",
        "UTF8Type": "STRING",
        "VarcharType": "STRING",
    }

    def execute(self, context: Context):
        hook = CassandraHook(cassandra_conn_id=self.cassandra_conn_id)

        query_extra = {}
        if self.query_timeout is not NOT_SET:
            query_extra["timeout"] = self.query_timeout

        cursor = hook.get_conn().execute(self.cql, **query_extra)

        # If a schema is set, create a BQ schema JSON file.
        if self.schema_filename:
            self.log.info("Writing local schema file")
            schema_file = self._write_local_schema_file(cursor)

            # Flush file before uploading
            schema_file["file_handle"].flush()

            self.log.info("Uploading schema file to GCS.")
            self._upload_to_gcs(schema_file)
            schema_file["file_handle"].close()

        counter = 0
        self.log.info("Writing local data files")
        for file_to_upload in self._write_local_data_files(cursor):
            # Flush file before uploading
            file_to_upload["file_handle"].flush()

            self.log.info("Uploading chunk file #%d to GCS.", counter)
            self._upload_to_gcs(file_to_upload)

            self.log.info("Removing local file")
            file_to_upload["file_handle"].close()
            counter += 1

        # Close all sessions and connection associated with this Cassandra cluster
        hook.shutdown_cluster()

    def _write_local_data_files(self, cursor):
        """
        Take a cursor, and writes results to a local file.

        :return: A dictionary where keys are filenames to be used as object
            names in GCS, and values are file handles to local files that
            contain the data for the GCS objects.
        """
        file_no = 0

        tmp_file_handle = NamedTemporaryFile(delete=True)
        file_to_upload = {
            "file_name": self.filename.format(file_no),
            "file_handle": tmp_file_handle,
        }
        for row in cursor:
            row_dict = self.generate_data_dict(row._fields, row)
            content = json.dumps(row_dict).encode("utf-8")
            tmp_file_handle.write(content)

            # Append newline to make dumps BigQuery compatible.
            tmp_file_handle.write(b"\n")

            if tmp_file_handle.tell() >= self.approx_max_file_size_bytes:
                file_no += 1

                yield file_to_upload
                tmp_file_handle = NamedTemporaryFile(delete=True)
                file_to_upload = {
                    "file_name": self.filename.format(file_no),
                    "file_handle": tmp_file_handle,
                }
        yield file_to_upload

    def _write_local_schema_file(self, cursor):
        """
        Take a cursor, and writes the BigQuery schema for the results to a local file system.

        :return: A dictionary where key is a filename to be used as an object
            name in GCS, and values are file handles to local files that
            contains the BigQuery schema fields in .json format.
        """
        schema = []
        tmp_schema_file_handle = NamedTemporaryFile(delete=True)

        for name, type_ in zip(cursor.column_names, cursor.column_types):
            schema.append(self.generate_schema_dict(name, type_))
        json_serialized_schema = json.dumps(schema).encode("utf-8")

        tmp_schema_file_handle.write(json_serialized_schema)
        schema_file_to_upload = {
            "file_name": self.schema_filename,
            "file_handle": tmp_schema_file_handle,
        }
        return schema_file_to_upload

    def _upload_to_gcs(self, file_to_upload):
        """Upload a file (data split or schema .json file) to Google Cloud Storage."""
        hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        hook.upload(
            bucket_name=self.bucket,
            object_name=file_to_upload.get("file_name"),
            filename=file_to_upload.get("file_handle").name,
            mime_type="application/json",
            gzip=self.gzip,
        )

    def generate_data_dict(self, names: Iterable[str], values: Any) -> dict[str, Any]:
        """Generate data structure that will be stored as file in GCS."""
        return {n: self.convert_value(v) for n, v in zip(names, values)}

    def convert_value(self, value: Any | None) -> Any | None:
        """Convert value to BQ type."""
        if not value or isinstance(value, (str, int, float, bool, dict)):
            return value
        if isinstance(value, bytes):
            return b64encode(value).decode("ascii")
        if isinstance(value, UUID):
            if self.encode_uuid:
                return b64encode(value.bytes).decode("ascii")
            return str(value)
        if isinstance(value, (datetime, Date)):
            return str(value)
        if isinstance(value, Decimal):
            return float(value)
        if isinstance(value, Time):
            return str(value).split(".")[0]
        if isinstance(value, (list, SortedSet)):
            return self.convert_array_types(value)
        if hasattr(value, "_fields"):
            return self.convert_user_type(value)
        if isinstance(value, tuple):
            return self.convert_tuple_type(value)
        if isinstance(value, OrderedMapSerializedKey):
            return self.convert_map_type(value)
        raise AirflowException(f"Unexpected value: {value}")

    def convert_array_types(self, value: list[Any] | SortedSet) -> list[Any]:
        """Map convert_value over array."""
        return [self.convert_value(nested_value) for nested_value in value]

    def convert_user_type(self, value: Any) -> dict[str, Any]:
        """
        Convert a user type to RECORD that contains n fields, where n is the number of attributes.

        Each element in the user type class will be converted to its corresponding data type in BQ.
        """
        names = value._fields
        values = [self.convert_value(getattr(value, name)) for name in names]
        return self.generate_data_dict(names, values)

    def convert_tuple_type(self, values: tuple[Any]) -> dict[str, Any]:
        """
        Convert a tuple to RECORD that contains n fields.

        Each field will be converted to its corresponding data type in bq and
        will be named 'field_<index>', where index is determined by the order
        of the tuple elements defined in cassandra.
        """
        names = [f"field_{i}" for i in range(len(values))]
        return self.generate_data_dict(names, values)

    def convert_map_type(self, value: OrderedMapSerializedKey) -> list[dict[str, Any]]:
        """
        Convert a map to a repeated RECORD that contains two fields: 'key' and 'value'.

        Each will be converted to its corresponding data type in BQ.
        """
        converted_map = []
        for k, v in zip(value.keys(), value.values()):
            converted_map.append({"key": self.convert_value(k), "value": self.convert_value(v)})
        return converted_map

    @classmethod
    def generate_schema_dict(cls, name: str, type_: Any) -> dict[str, Any]:
        """Generate BQ schema."""
        field_schema: dict[str, Any] = {}
        field_schema.update({"name": name})
        field_schema.update({"type_": cls.get_bq_type(type_)})
        field_schema.update({"mode": cls.get_bq_mode(type_)})
        fields = cls.get_bq_fields(type_)
        if fields:
            field_schema.update({"fields": fields})
        return field_schema

    @classmethod
    def get_bq_fields(cls, type_: Any) -> list[dict[str, Any]]:
        """Convert non simple type value to BQ representation."""
        if cls.is_simple_type(type_):
            return []

        # In case of not simple type
        names: list[str] = []
        types: list[Any] = []
        if cls.is_array_type(type_) and cls.is_record_type(type_.subtypes[0]):
            names = type_.subtypes[0].fieldnames
            types = type_.subtypes[0].subtypes
        elif cls.is_record_type(type_):
            names = type_.fieldnames
            types = type_.subtypes

        if types and not names and type_.cassname == "TupleType":
            names = [f"field_{i}" for i in range(len(types))]
        elif types and not names and type_.cassname == "MapType":
            names = ["key", "value"]

        return [cls.generate_schema_dict(n, t) for n, t in zip(names, types)]

    @staticmethod
    def is_simple_type(type_: Any) -> bool:
        """Check if type is a simple type."""
        return type_.cassname in CassandraToGCSOperator.CQL_TYPE_MAP

    @staticmethod
    def is_array_type(type_: Any) -> bool:
        """Check if type is an array type."""
        return type_.cassname in ["ListType", "SetType"]

    @staticmethod
    def is_record_type(type_: Any) -> bool:
        """Check the record type."""
        return type_.cassname in ["UserType", "TupleType", "MapType"]

    @classmethod
    def get_bq_type(cls, type_: Any) -> str:
        """Convert type to equivalent BQ type."""
        if cls.is_simple_type(type_):
            return CassandraToGCSOperator.CQL_TYPE_MAP[type_.cassname]
        if cls.is_record_type(type_):
            return "RECORD"
        if cls.is_array_type(type_):
            return cls.get_bq_type(type_.subtypes[0])
        raise AirflowException("Not a supported type_: " + type_.cassname)

    @classmethod
    def get_bq_mode(cls, type_: Any) -> str:
        """Convert type to equivalent BQ mode."""
        if cls.is_array_type(type_) or type_.cassname == "MapType":
            return "REPEATED"
        if cls.is_record_type(type_) or cls.is_simple_type(type_):
            return "NULLABLE"
        raise AirflowException("Not a supported type_: " + type_.cassname)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import urllib.parse

import pytest

from airflow.providers.common.compat.assets import Asset
from airflow.providers.google.assets.gcs import convert_asset_to_openlineage, create_asset, sanitize_uri


def test_sanitize_uri():
    uri = sanitize_uri(urllib.parse.urlsplit("gs://bucket/dir/file.txt"))
    result = sanitize_uri(uri)
    assert result.scheme == "gs"
    assert result.netloc == "bucket"
    assert result.path == "/dir/file.txt"


def test_sanitize_uri_no_netloc():
    with pytest.raises(ValueError, match="URI format gs:// must contain"):
        sanitize_uri(urllib.parse.urlsplit("gs://"))


def test_sanitize_uri_no_path():
    uri = sanitize_uri(urllib.parse.urlsplit("gs://bucket"))
    result = sanitize_uri(uri)
    assert result.scheme == "gs"
    assert result.netloc == "bucket"
    assert result.path == ""


def test_create_asset():
    assert create_asset(bucket="test-bucket", key="test-path") == Asset(uri="gs://test-bucket/test-path")
    assert create_asset(bucket="test-bucket", key="test-dir/test-path") == Asset(
        uri="gs://test-bucket/test-dir/test-path"
    )


def test_sanitize_uri_trailing_slash():
    uri = sanitize_uri(urllib.parse.urlsplit("gs://bucket/"))
    result = sanitize_uri(uri)
    assert result.scheme == "gs"
    assert result.netloc == "bucket"
    assert result.path == "/"


def test_convert_asset_to_openlineage_valid():
    uri = "gs://bucket/dir/file.txt"
    ol_dataset = convert_asset_to_openlineage(asset=Asset(uri=uri), lineage_context=None)
    assert ol_dataset.namespace == "gs://bucket"
    assert ol_dataset.name == "dir/file.txt"


@pytest.mark.parametrize("uri", ("gs://bucket", "gs://bucket/"))
def test_convert_asset_to_openlineage_no_path(uri):
    ol_dataset = convert_asset_to_openlineage(asset=Asset(uri=uri), lineage_context=None)
    assert ol_dataset.namespace == "gs://bucket"
    assert ol_dataset.name == "/"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a mechanism for providing temporary Google Cloud authentication."""

from __future__ import annotations

import json
import logging
import os
import tempfile
from collections.abc import Collection, Generator, Sequence
from contextlib import ExitStack, contextmanager
from urllib.parse import urlencode

import google.auth
import google.oauth2.service_account
from google.auth import impersonated_credentials
from google.auth.credentials import AnonymousCredentials, Credentials
from google.auth.environment_vars import CREDENTIALS, LEGACY_PROJECT, PROJECT

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud._internal_client.secret_manager_client import _SecretManagerClient
from airflow.providers.google.cloud.utils.external_token_supplier import (
    ClientCredentialsGrantFlowTokenSupplier,
)
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.process_utils import patch_environ

log = logging.getLogger(__name__)

AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT = "AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT"
_DEFAULT_SCOPES: Sequence[str] = ("https://www.googleapis.com/auth/cloud-platform",)


def build_gcp_conn(
    key_file_path: str | None = None,
    scopes: Sequence[str] | None = None,
    project_id: str | None = None,
) -> str:
    """
    Build a uri that can be used as :envvar:`AIRFLOW_CONN_{CONN_ID}` with provided values.

    :param key_file_path: Path to service key.
    :param scopes: Required OAuth scopes.
    :param project_id: The Google Cloud project id to be used for the connection.
    :return: String representing Airflow connection.
    """
    conn = "google-cloud-platform://?{}"
    query_params = {}
    if key_file_path:
        query_params["key_path"] = key_file_path
    if scopes:
        scopes_string = ",".join(scopes)
        query_params["scope"] = scopes_string
    if project_id:
        query_params["projects"] = project_id

    query = urlencode(query_params)
    return conn.format(query)


@contextmanager
def provide_gcp_credentials(
    key_file_path: str | None = None,
    key_file_dict: dict | None = None,
) -> Generator[None, None, None]:
    """
    Context manager that provides Google Cloud credentials for Application Default Credentials (ADC).

    .. seealso::
        `Application Default Credentials (ADC) strategy`__.

    It can be used to provide credentials for external programs (e.g. gcloud) that expect authorization
    file in ``GOOGLE_APPLICATION_CREDENTIALS`` environment variable.

    :param key_file_path: Path to file with Google Cloud Service Account .json file.
    :param key_file_dict: Dictionary with credentials.

    __ https://cloud.google.com/docs/authentication/production
    """
    if not key_file_path and not key_file_dict:
        raise ValueError("Please provide `key_file_path` or `key_file_dict`.")

    if key_file_path and key_file_path.endswith(".p12"):
        raise AirflowException("Legacy P12 key file are not supported, use a JSON key file.")

    with tempfile.NamedTemporaryFile(mode="w+t") as conf_file:
        if not key_file_path and key_file_dict:
            conf_file.write(json.dumps(key_file_dict))
            conf_file.flush()
            key_file_path = conf_file.name
        if key_file_path:
            with patch_environ({CREDENTIALS: key_file_path}):
                yield
        else:
            # We will use the default service account credentials.
            yield


@contextmanager
def provide_gcp_connection(
    key_file_path: str | None = None,
    scopes: Sequence | None = None,
    project_id: str | None = None,
) -> Generator[None, None, None]:
    """
    Context manager that provides a temporary value of :envvar:`AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT` connection.

    It builds a new connection that includes path to provided service json, required scopes and project id.

    :param key_file_path: Path to file with Google Cloud Service Account .json file.
    :param scopes: OAuth scopes for the connection
    :param project_id: The id of Google Cloud project for the connection.
    """
    if key_file_path and key_file_path.endswith(".p12"):
        raise AirflowException("Legacy P12 key file are not supported, use a JSON key file.")

    conn = build_gcp_conn(scopes=scopes, key_file_path=key_file_path, project_id=project_id)

    with patch_environ({AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT: conn}):
        yield


@contextmanager
def provide_gcp_conn_and_credentials(
    key_file_path: str | None = None,
    scopes: Sequence | None = None,
    project_id: str | None = None,
) -> Generator[None, None, None]:
    """
    Context manager that provides GPC connection and credentials.

     It provides both:

    - Google Cloud credentials for application supporting `Application Default Credentials (ADC)
      strategy`__.
    - temporary value of :envvar:`AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT` connection

    :param key_file_path: Path to file with Google Cloud Service Account .json file.
    :param scopes: OAuth scopes for the connection
    :param project_id: The id of Google Cloud project for the connection.

    __ https://cloud.google.com/docs/authentication/production
    """
    with ExitStack() as stack:
        if key_file_path:
            stack.enter_context(provide_gcp_credentials(key_file_path))  # type; ignore
        if project_id:
            stack.enter_context(  # type; ignore
                patch_environ({PROJECT: project_id, LEGACY_PROJECT: project_id})
            )

        stack.enter_context(provide_gcp_connection(key_file_path, scopes, project_id))  # type; ignore
        yield


class _CredentialProvider(LoggingMixin):
    """
    Prepare the Credentials object for Google API and the associated project_id.

    Only either `key_path` or `keyfile_dict` should be provided, or an exception will
    occur. If neither of them are provided, return default credentials for the current environment

    :param key_path: Path to Google Cloud Service Account key file (JSON).
    :param keyfile_dict: A dict representing Cloud Service Account as in the Credential JSON file
    :param key_secret_name: Keyfile Secret Name in GCP Secret Manager.
    :param key_secret_project_id: Project ID to read the secrets from. If not passed, the project ID from
        default credentials will be used.
    :param credential_config_file: File path to or content of a GCP credential configuration file.
    :param scopes:  OAuth scopes for the connection
    :param delegate_to: The account to impersonate using domain-wide delegation of authority,
        if any. For this to work, the service account making the request must have
        domain-wide delegation enabled.
    :param disable_logging: If true, disable all log messages, which allows you to use this
        class to configure Logger.
    :param target_principal: The service account to directly impersonate using short-term
        credentials, if any. For this to work, the target_principal account must grant
        the originating account the Service Account Token Creator IAM role.
    :param delegates: optional chained list of accounts required to get the access_token of
        target_principal. If set, the sequence of identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account and target_principal
        granting the role to the last account from the list.
    :param is_anonymous: Provides an anonymous set of credentials,
        which is useful for APIs which do not require authentication.
    """

    def __init__(
        self,
        key_path: str | None = None,
        keyfile_dict: dict[str, str] | None = None,
        credential_config_file: dict[str, str] | str | None = None,
        key_secret_name: str | None = None,
        key_secret_project_id: str | None = None,
        scopes: Collection[str] | None = None,
        delegate_to: str | None = None,
        disable_logging: bool = False,
        target_principal: str | None = None,
        delegates: Sequence[str] | None = None,
        is_anonymous: bool | None = None,
        idp_issuer_url: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        idp_extra_params_dict: dict[str, str] | None = None,
    ) -> None:
        super().__init__()
        key_options_map = {
            "key_path": key_path,
            "keyfile_dict": keyfile_dict,
            "credential_config_file": credential_config_file,
            "key_secret_name": key_secret_name,
            "is_anonymous": is_anonymous,
        }
        key_options_label_provided = [label for label, credential in key_options_map.items() if credential]
        if len(key_options_label_provided) > 1:
            raise AirflowException(
                f"The `keyfile_dict`, `key_path`, `credential_config_file`, `is_anonymous` and"
                f" `key_secret_name` fields are all mutually exclusive. "
                f"Received options: {key_options_label_provided}. Please provide only one value."
            )
        self.key_path = key_path
        self.keyfile_dict = keyfile_dict
        self.credential_config_file = credential_config_file
        self.key_secret_name = key_secret_name
        self.key_secret_project_id = key_secret_project_id
        self.scopes = scopes
        self.delegate_to = delegate_to
        self.disable_logging = disable_logging
        self.target_principal = target_principal
        self.delegates = delegates
        self.is_anonymous = is_anonymous
        self.idp_issuer_url = idp_issuer_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.idp_extra_params_dict = idp_extra_params_dict

    def get_credentials_and_project(self) -> tuple[Credentials, str]:
        """
        Get current credentials and project ID.

        Project ID is an empty string when using anonymous credentials.

        :return: Google Auth Credentials
        """
        if self.is_anonymous:
            credentials: Credentials = AnonymousCredentials()
            project_id = ""
        else:
            if self.key_path:
                credentials, project_id = self._get_credentials_using_key_path()
            elif self.key_secret_name:
                credentials, project_id = self._get_credentials_using_key_secret_name()
            elif self.keyfile_dict:
                credentials, project_id = self._get_credentials_using_keyfile_dict()
            elif self.idp_issuer_url:
                credentials, project_id = (
                    self._get_credentials_using_credential_config_file_and_token_supplier()
                )
            elif self.credential_config_file:
                credentials, project_id = self._get_credentials_using_credential_config_file()
            else:
                credentials, project_id = self._get_credentials_using_adc()
            if self.delegate_to:
                if hasattr(credentials, "with_subject"):
                    credentials = credentials.with_subject(self.delegate_to)
                else:
                    raise AirflowException(
                        "The `delegate_to` parameter cannot be used here as the current "
                        "authentication method does not support account impersonate. "
                        "Please use service-account for authorization."
                    )

            if self.target_principal:
                credentials = impersonated_credentials.Credentials(
                    source_credentials=credentials,
                    target_principal=self.target_principal,
                    delegates=self.delegates,
                    target_scopes=self.scopes,
                )

                project_id = _get_project_id_from_service_account_email(self.target_principal)

        return credentials, project_id

    def _get_credentials_using_keyfile_dict(self) -> tuple[Credentials, str]:
        self._log_debug("Getting connection using JSON Dict")
        # Depending on how the JSON was formatted, it may contain
        # escaped newlines. Convert those to actual newlines.
        if self.keyfile_dict is None:
            raise ValueError("The keyfile_dict field is None, and we need it for keyfile_dict auth.")
        self.keyfile_dict["private_key"] = self.keyfile_dict["private_key"].replace("\\n", "\n")
        credentials = google.oauth2.service_account.Credentials.from_service_account_info(
            self.keyfile_dict, scopes=self.scopes
        )
        project_id = credentials.project_id
        return credentials, project_id

    def _get_credentials_using_key_path(self) -> tuple[Credentials, str]:
        if self.key_path is None:
            raise ValueError("The ky_path field is None, and we need it for keyfile_dict auth.")
        if self.key_path.endswith(".p12"):
            raise AirflowException("Legacy P12 key file are not supported, use a JSON key file.")

        if not self.key_path.endswith(".json"):
            raise AirflowException("Unrecognised extension for key file.")

        self._log_debug("Getting connection using JSON key file %s", self.key_path)
        credentials = google.oauth2.service_account.Credentials.from_service_account_file(
            self.key_path, scopes=self.scopes
        )
        project_id = credentials.project_id
        return credentials, project_id

    def _get_credentials_using_key_secret_name(self) -> tuple[Credentials, str]:
        self._log_debug("Getting connection using JSON key data from GCP secret: %s", self.key_secret_name)

        # Use ADC to access GCP Secret Manager.
        adc_credentials, adc_project_id = google.auth.default(scopes=self.scopes)
        secret_manager_client = _SecretManagerClient(credentials=adc_credentials)

        if self.key_secret_name is None:
            raise ValueError("The key_secret_name field is None, and we need it for keyfile_dict auth.")
        if not secret_manager_client.is_valid_secret_name(self.key_secret_name):
            raise AirflowException("Invalid secret name specified for fetching JSON key data.")

        secret_value = secret_manager_client.get_secret(
            secret_id=self.key_secret_name,
            project_id=self.key_secret_project_id if self.key_secret_project_id else adc_project_id,
        )
        if secret_value is None:
            raise AirflowException(f"Failed getting value of secret {self.key_secret_name}.")

        try:
            keyfile_dict = json.loads(secret_value)
        except json.decoder.JSONDecodeError:
            raise AirflowException("Key data read from GCP Secret Manager is not valid JSON.")

        credentials = google.oauth2.service_account.Credentials.from_service_account_info(
            keyfile_dict, scopes=self.scopes
        )
        project_id = credentials.project_id
        return credentials, project_id

    def _get_credentials_using_credential_config_file(self) -> tuple[Credentials, str]:
        if isinstance(self.credential_config_file, str) and os.path.exists(self.credential_config_file):
            self._log_info(
                f"Getting connection using credential configuration file: `{self.credential_config_file}`"
            )
            credentials, project_id = google.auth.load_credentials_from_file(
                self.credential_config_file, scopes=self.scopes
            )
        else:
            with tempfile.NamedTemporaryFile(mode="w+t") as temp_credentials_fd:
                if isinstance(self.credential_config_file, dict):
                    self._log_info("Getting connection using credential configuration dict.")
                    temp_credentials_fd.write(json.dumps(self.credential_config_file))
                elif isinstance(self.credential_config_file, str):
                    self._log_info("Getting connection using credential configuration string.")
                    temp_credentials_fd.write(self.credential_config_file)

                temp_credentials_fd.flush()
                credentials, project_id = google.auth.load_credentials_from_file(
                    temp_credentials_fd.name, scopes=self.scopes
                )

        return credentials, project_id

    def _get_credentials_using_credential_config_file_and_token_supplier(self):
        self._log_info(
            "Getting connection using credential configuration file and external Identity Provider."
        )

        if not self.credential_config_file:
            raise AirflowException(
                "Credential Configuration File is needed to use authentication by External Identity Provider."
            )

        info = _get_info_from_credential_configuration_file(self.credential_config_file)
        info["subject_token_supplier"] = ClientCredentialsGrantFlowTokenSupplier(
            oidc_issuer_url=self.idp_issuer_url, client_id=self.client_id, client_secret=self.client_secret
        )

        credentials, project_id = google.auth.load_credentials_from_dict(info=info, scopes=self.scopes)
        return credentials, project_id

    def _get_credentials_using_adc(self) -> tuple[Credentials, str]:
        self._log_info(
            "Getting connection using `google.auth.default()` since no explicit credentials are provided."
        )
        credentials, project_id = google.auth.default(scopes=self.scopes)
        return credentials, project_id

    def _log_info(self, *args, **kwargs) -> None:
        if not self.disable_logging:
            self.log.info(*args, **kwargs)

    def _log_debug(self, *args, **kwargs) -> None:
        if not self.disable_logging:
            self.log.debug(*args, **kwargs)


def get_credentials_and_project_id(*args, **kwargs) -> tuple[Credentials, str]:
    """Return the Credentials object for Google API and the associated project_id."""
    return _CredentialProvider(*args, **kwargs).get_credentials_and_project()


def _get_scopes(scopes: str | None = None) -> Sequence[str]:
    """
    Parse a comma-separated string containing OAuth2 scopes if `scopes` is provided; otherwise return default.

    :param scopes: A comma-separated string containing OAuth2 scopes
    :return: Returns the scope defined in the connection configuration, or the default scope
    """
    return [s.strip() for s in scopes.split(",")] if scopes else _DEFAULT_SCOPES


def _get_target_principal_and_delegates(
    impersonation_chain: str | Sequence[str] | None = None,
) -> tuple[str | None, Sequence[str] | None]:
    """
    Get the target_principal and optional list of delegates from impersonation_chain.

    Analyze contents of impersonation_chain and return target_principal (the service account
    to directly impersonate using short-term credentials, if any) and optional list of delegates
    required to get the access_token of target_principal.

    :param impersonation_chain: the service account to impersonate or a chained list leading to this
        account

    :return: Returns the tuple of target_principal and delegates
    """
    if not impersonation_chain:
        return None, None

    if isinstance(impersonation_chain, str):
        return impersonation_chain, None

    return impersonation_chain[-1], impersonation_chain[:-1]


def _get_project_id_from_service_account_email(service_account_email: str) -> str:
    """
    Extract project_id from service account's email address.

    :param service_account_email: email of the service account.

    :return: Returns the project_id of the provided service account.
    """
    try:
        return service_account_email.split("@")[1].split(".")[0]
    except IndexError:
        raise AirflowException(
            f"Could not extract project_id from service account's email: {service_account_email}."
        )


def _get_info_from_credential_configuration_file(
    credential_configuration_file: str | dict[str, str],
) -> dict[str, str]:
    """
    Extract the Credential Configuration File information, either from a json file, json string or dictionary.

    :param credential_configuration_file: File path or content (as json string or dictionary) of a GCP credential configuration file.

    :return: Returns a dictionary containing the Credential Configuration File information.
    """
    # if it's already a dict, just return it
    if isinstance(credential_configuration_file, dict):
        return credential_configuration_file

    if not isinstance(credential_configuration_file, str):
        raise AirflowException(
            f"Invalid argument type, expected str or dict, got {type(credential_configuration_file)}."
        )

    if os.path.exists(credential_configuration_file):  # attempts to load from json file
        with open(credential_configuration_file) as file_obj:
            try:
                return json.load(file_obj)
            except ValueError:
                raise AirflowException(
                    f"Credential Configuration File '{credential_configuration_file}' is not a valid json file."
                )

    # if not a file, attempt to load it from a json string
    try:
        return json.loads(credential_configuration_file)
    except ValueError:
        raise AirflowException("Credential Configuration File is not a valid json string.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for Google Vertex AI Generative Model Tuning Tasks.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

import requests

from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.local_to_gcs import LocalFilesystemToGCSOperator

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]
from google.genai.types import TuningDataset

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gen_ai import (
    GenAISupervisedFineTuningTrainOperator,
)
from airflow.providers.google.common.utils.get_secret import get_secret


def _get_actual_model(key) -> str:
    source_model: str | None = None
    try:
        response = requests.get("https://generativelanguage.googleapis.com/v1/models", {"key": key})
        response.raise_for_status()
        available_models = response.json()
    except requests.exceptions.RequestException as e:
        print(f"Error fetching models from API: {e}")
        return ""
    for model in available_models.get("models", []):
        try:
            model_name = model["name"].split("/")[-1]
            splited_model_name = model_name.split("-")
            if not source_model and "flash" in model_name:
                source_model = model_name
            elif (
                source_model
                and "flash" in model_name
                and float(source_model.split("-")[1]) < float(splited_model_name[1])
            ):
                source_model = model_name
            elif (
                source_model
                and "flash" in model_name
                and (
                    float(source_model.split("-")[1]) == float(splited_model_name[1])
                    and int(splited_model_name[-1]) > int(source_model.split("-")[-1])
                )
            ):
                source_model = model_name
        except (ValueError, IndexError) as e:
            print(f"Could not parse model name '{model.get('name')}'. Skipping. Error: {e}")
            continue
    if not source_model:
        raise ValueError("Source model not found")
    return source_model


PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "gen_ai_generative_model_tuning_dag"
REGION = "us-central1"
GEMINI_API_KEY = "api_key"
SOURCE_MODEL = "{{ task_instance.xcom_pull('get_actual_model') }}"
TRAIN_DATASET = TuningDataset(
    gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
)
TUNED_MODEL_DISPLAY_NAME = "my_tuned_gemini_model"
TUNING_JOB_CONFIG = {"tuned_model_display_name": TUNED_MODEL_DISPLAY_NAME}
TUNED_VIDEO_MODEL_DISPLAY_NAME = "my_tuned_gemini_video_model"
TUNING_JOB_VIDEO_MODEL_CONFIG = {"tuned_model_display_name": TUNED_VIDEO_MODEL_DISPLAY_NAME}

BUCKET_NAME = f"bucket_tuning_dag_{PROJECT_ID}"
FILE_NAME = "video_tuning_dataset.jsonl"
UPLOAD_FILE_PATH = str(Path(__file__).parent / "resources" / FILE_NAME)
TRAIN_VIDEO_DATASET = TuningDataset(gcs_uri=f"gs://{BUCKET_NAME}/{FILE_NAME}")


with DAG(
    dag_id=DAG_ID,
    description="Sample DAG with generative model tuning tasks.",
    schedule="@once",
    start_date=datetime(2024, 1, 1),
    catchup=False,
    tags=["example", "vertex_ai", "generative_model"],
    render_template_as_native_obj=True,
) as dag:

    @task
    def get_gemini_api_key():
        return get_secret(GEMINI_API_KEY)

    get_gemini_api_key_task = get_gemini_api_key()

    @task
    def get_actual_model(key):
        return _get_actual_model(key)

    get_actual_model_task = get_actual_model(get_gemini_api_key_task)

    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=BUCKET_NAME,
        project_id=PROJECT_ID,
    )

    upload_file = LocalFilesystemToGCSOperator(
        task_id="upload_file",
        src=UPLOAD_FILE_PATH,
        dst=FILE_NAME,
        bucket=BUCKET_NAME,
    )

    delete_bucket = GCSDeleteBucketOperator(task_id="delete_bucket", bucket_name=BUCKET_NAME)

    # [START how_to_cloud_gen_ai_supervised_fine_tuning_train_operator]
    sft_train_task = GenAISupervisedFineTuningTrainOperator(
        task_id="sft_train_task",
        project_id=PROJECT_ID,
        location=REGION,
        source_model=SOURCE_MODEL,
        training_dataset=TRAIN_DATASET,
        tuning_job_config=TUNING_JOB_CONFIG,
    )
    # [END how_to_cloud_gen_ai_supervised_fine_tuning_train_operator]

    # [START how_to_cloud_gen_ai_supervised_fine_tuning_train_operator_for_video]
    sft_video_task = GenAISupervisedFineTuningTrainOperator(
        task_id="sft_train_video_task",
        project_id=PROJECT_ID,
        location=REGION,
        source_model=SOURCE_MODEL,
        training_dataset=TRAIN_VIDEO_DATASET,
        tuning_job_config=TUNING_JOB_VIDEO_MODEL_CONFIG,
    )
    # [END how_to_cloud_gen_ai_supervised_fine_tuning_train_operator_for_video]

    delete_bucket.trigger_rule = TriggerRule.ALL_DONE

    (
        get_gemini_api_key_task
        >> get_actual_model_task
        >> create_bucket
        >> upload_file
        >> [sft_train_task, sft_video_task]
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that demonstrates interactions with Google Cloud Transfer.
"""

from __future__ import annotations

import os
from copy import deepcopy
from datetime import datetime, timedelta, timezone

from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.transfers.gcs_to_s3 import GCSToS3Operator
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import (
    ALREADY_EXISTING_IN_SINK,
    AWS_S3_DATA_SOURCE,
    BUCKET_NAME,
    DESCRIPTION,
    FILTER_JOB_NAMES,
    FILTER_PROJECT_ID,
    GCS_DATA_SINK,
    JOB_NAME,
    PROJECT_ID,
    SCHEDULE,
    SCHEDULE_END_DATE,
    SCHEDULE_START_DATE,
    START_TIME_OF_DAY,
    STATUS,
    TRANSFER_OPTIONS,
    TRANSFER_SPEC,
    GcpTransferJobsStatus,
    GcpTransferOperationStatus,
)
from airflow.providers.google.cloud.operators.cloud_storage_transfer_service import (
    CloudDataTransferServiceCancelOperationOperator,
    CloudDataTransferServiceCreateJobOperator,
    CloudDataTransferServiceDeleteJobOperator,
    CloudDataTransferServiceGetOperationOperator,
    CloudDataTransferServiceListOperationsOperator,
    CloudDataTransferServicePauseOperationOperator,
    CloudDataTransferServiceResumeOperationOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.cloud_storage_transfer_service import (
    CloudDataTransferServiceJobStatusSensor,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
GCP_PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "example_gcp_transfer_aws"

EXAMPLE_BUCKET = "airflow-system-tests-resources"
EXAMPLE_FILE = "storage-transfer/big_file.dat"
BUCKET_SOURCE_AWS = f"bucket-aws-{DAG_ID}-{ENV_ID}".replace("_", "-")
BUCKET_TARGET_GCS = f"bucket-gcs-{DAG_ID}-{ENV_ID}".replace("_", "-")
WAIT_FOR_OPERATION_POKE_INTERVAL = int(os.environ.get("WAIT_FOR_OPERATION_POKE_INTERVAL", str(5)))

GCP_DESCRIPTION = "description"
GCP_TRANSFER_JOB_NAME = f"transferJobs/sampleJob-{DAG_ID}-{ENV_ID}".replace("-", "_")
GCP_TRANSFER_JOB_2_NAME = f"transferJobs/sampleJob2-{DAG_ID}-{ENV_ID}".replace("-", "_")

# [START howto_operator_gcp_transfer_create_job_body_aws]
aws_to_gcs_transfer_body = {
    DESCRIPTION: GCP_DESCRIPTION,
    STATUS: GcpTransferJobsStatus.ENABLED,
    PROJECT_ID: GCP_PROJECT_ID,
    JOB_NAME: GCP_TRANSFER_JOB_NAME,
    SCHEDULE: {
        SCHEDULE_START_DATE: datetime(2015, 1, 1).date(),
        SCHEDULE_END_DATE: datetime(2030, 1, 1).date(),
        START_TIME_OF_DAY: (datetime.now(tz=timezone.utc) + timedelta(minutes=1)).time(),
    },
    TRANSFER_SPEC: {
        AWS_S3_DATA_SOURCE: {BUCKET_NAME: BUCKET_SOURCE_AWS},
        GCS_DATA_SINK: {BUCKET_NAME: BUCKET_TARGET_GCS},
        TRANSFER_OPTIONS: {ALREADY_EXISTING_IN_SINK: True},
    },
}
# [END howto_operator_gcp_transfer_create_job_body_aws]

aws_to_gcs_transfer_body_2 = deepcopy(aws_to_gcs_transfer_body)
aws_to_gcs_transfer_body_2[JOB_NAME] = GCP_TRANSFER_JOB_2_NAME

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example", "aws", "gcs", "transfer"],
) as dag:
    create_bucket_s3 = S3CreateBucketOperator(
        task_id="create_bucket_s3", bucket_name=BUCKET_SOURCE_AWS, region_name="us-east-1"
    )

    upload_file_to_s3 = GCSToS3Operator(
        task_id="upload_file_to_s3",
        gcp_user_project=GCP_PROJECT_ID,
        gcs_bucket=EXAMPLE_BUCKET,
        prefix=EXAMPLE_FILE,
        dest_s3_key=f"s3://{BUCKET_SOURCE_AWS}",
        replace=True,
    )
    #
    create_bucket_gcs = GCSCreateBucketOperator(
        task_id="create_bucket_gcs",
        bucket_name=BUCKET_TARGET_GCS,
        project_id=GCP_PROJECT_ID,
    )

    # [START howto_operator_gcp_transfer_create_job]
    create_transfer_job_s3_to_gcs = CloudDataTransferServiceCreateJobOperator(
        task_id="create_transfer_job_s3_to_gcs", body=aws_to_gcs_transfer_body
    )
    # [END howto_operator_gcp_transfer_create_job]

    wait_for_operation_to_start = CloudDataTransferServiceJobStatusSensor(
        task_id="wait_for_operation_to_start",
        job_name="{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs')['name']}}",
        project_id=GCP_PROJECT_ID,
        expected_statuses={GcpTransferOperationStatus.IN_PROGRESS},
        poke_interval=WAIT_FOR_OPERATION_POKE_INTERVAL,
    )

    # [START howto_operator_gcp_transfer_pause_operation]
    pause_operation = CloudDataTransferServicePauseOperationOperator(
        task_id="pause_operation",
        operation_name="{{task_instance.xcom_pull('wait_for_operation_to_start', "
        "key='sensed_operations')[0]['name']}}",
    )
    # [END howto_operator_gcp_transfer_pause_operation]

    # [START howto_operator_gcp_transfer_list_operations]
    list_operations = CloudDataTransferServiceListOperationsOperator(
        task_id="list_operations",
        request_filter={
            FILTER_PROJECT_ID: GCP_PROJECT_ID,
            FILTER_JOB_NAMES: ["{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs')['name']}}"],
        },
    )
    # [END howto_operator_gcp_transfer_list_operations]

    # [START howto_operator_gcp_transfer_get_operation]
    get_operation = CloudDataTransferServiceGetOperationOperator(
        task_id="get_operation", operation_name="{{task_instance.xcom_pull('list_operations')[0]['name']}}"
    )
    # [END howto_operator_gcp_transfer_get_operation]

    # [START howto_operator_gcp_transfer_resume_operation]
    resume_operation = CloudDataTransferServiceResumeOperationOperator(
        task_id="resume_operation", operation_name="{{task_instance.xcom_pull('get_operation')['name']}}"
    )
    # [END howto_operator_gcp_transfer_resume_operation]

    # [START howto_operator_gcp_transfer_wait_operation]
    wait_for_operation_to_end = CloudDataTransferServiceJobStatusSensor(
        task_id="wait_for_operation_to_end",
        job_name="{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs')['name']}}",
        project_id=GCP_PROJECT_ID,
        expected_statuses={GcpTransferOperationStatus.SUCCESS},
        poke_interval=WAIT_FOR_OPERATION_POKE_INTERVAL,
    )
    # [END howto_operator_gcp_transfer_wait_operation]

    create_second_transfer_job_from_aws = CloudDataTransferServiceCreateJobOperator(
        task_id="create_transfer_job_s3_to_gcs_2", body=aws_to_gcs_transfer_body_2
    )

    wait_for_operation_to_start_2 = CloudDataTransferServiceJobStatusSensor(
        task_id="wait_for_operation_to_start_2",
        job_name="{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs_2')['name']}}",
        project_id=GCP_PROJECT_ID,
        expected_statuses={GcpTransferOperationStatus.IN_PROGRESS},
        poke_interval=WAIT_FOR_OPERATION_POKE_INTERVAL,
    )

    # [START howto_operator_gcp_transfer_cancel_operation]
    cancel_operation = CloudDataTransferServiceCancelOperationOperator(
        task_id="cancel_operation",
        operation_name="{{task_instance.xcom_pull("
        "'wait_for_operation_to_start_2', key='sensed_operations')[0]['name']}}",
    )
    # [END howto_operator_gcp_transfer_cancel_operation]

    # [START howto_operator_gcp_transfer_delete_job]
    delete_transfer_job_s3_to_gcs = CloudDataTransferServiceDeleteJobOperator(
        task_id="delete_transfer_job_s3_to_gcs",
        job_name="{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs')['name']}}",
        project_id=GCP_PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_gcp_transfer_delete_job]

    delete_transfer_job_s3_to_gcs_2 = CloudDataTransferServiceDeleteJobOperator(
        task_id="delete_transfer_job_s3_to_gcs_2",
        job_name="{{task_instance.xcom_pull('create_transfer_job_s3_to_gcs_2')['name']}}",
        project_id=GCP_PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket_s3 = S3DeleteBucketOperator(
        task_id="delete_bucket_s3",
        bucket_name=BUCKET_SOURCE_AWS,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket_gcs = GCSDeleteBucketOperator(
        task_id="delete_bucket_gcs",
        bucket_name=BUCKET_TARGET_GCS,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        [create_bucket_s3 >> upload_file_to_s3, create_bucket_gcs]
        # TEST BODY
        >> create_transfer_job_s3_to_gcs
        >> wait_for_operation_to_start
        >> pause_operation
        >> list_operations
        >> get_operation
        >> resume_operation
        >> wait_for_operation_to_end
        >> create_second_transfer_job_from_aws
        >> wait_for_operation_to_start_2
        >> cancel_operation
        # TEST TEARDOWN
        >> [
            delete_transfer_job_s3_to_gcs,
            delete_transfer_job_s3_to_gcs_2,
            delete_bucket_gcs,
            delete_bucket_s3,
        ]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google BigQuery to MySQL operator."""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from functools import cached_property

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.google.cloud.transfers.bigquery_to_sql import BigQueryToSqlBaseOperator
from airflow.providers.mysql.hooks.mysql import MySqlHook


class BigQueryToMySqlOperator(BigQueryToSqlBaseOperator):
    """
    Fetch data from a BigQuery table (alternatively fetch selected columns) and insert it into a MySQL table.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:BigQueryToMySqlOperator`

    :param mysql_table: target MySQL table, use dot notation to target a
        specific database. It is deprecated: use target_table_name instead. (templated)
    :param target_table_name: target MySQL table. It takes precedence over mysql_table. (templated)
    :param mysql_conn_id: Reference to :ref:`mysql connection id <howto/connection:mysql>`.

    .. warning::
        The `mysql_table` parameter has been deprecated. Use `target_table_name` instead.
    """

    template_fields: Sequence[str] = (*BigQueryToSqlBaseOperator.template_fields, "dataset_id", "table_id")

    def __init__(
        self,
        *,
        mysql_table: str | None = None,
        target_table_name: str | None = None,
        mysql_conn_id: str = "mysql_default",
        dataset_id: str | None = None,
        table_id: str | None = None,
        **kwargs,
    ) -> None:
        if mysql_table is not None:
            warnings.warn(
                "The `mysql_table` parameter has been deprecated. Use `target_table_name` instead.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )

            if target_table_name is not None:
                raise ValueError(
                    f"Cannot set both arguments: mysql_table={mysql_table!r} and "
                    f"target_table_name={target_table_name!r}."
                )

            target_table_name = mysql_table

        super().__init__(
            target_table_name=target_table_name, dataset_id=dataset_id, table_id=table_id, **kwargs
        )
        self.mysql_conn_id = mysql_conn_id

    @cached_property
    def mysql_hook(self) -> MySqlHook:
        return MySqlHook(schema=self.database, mysql_conn_id=self.mysql_conn_id)

    def get_sql_hook(self) -> MySqlHook:
        return self.mysql_hook

    def execute(self, context):
        # Set source_project_dataset_table here, after hooks are initialized and project_id is available
        project_id = self.bigquery_hook.project_id
        self.source_project_dataset_table = f"{project_id}.{self.dataset_id}.{self.table_id}"
        return super().execute(context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from google.cloud.bigquery.table import Row, RowIterator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
    from airflow.sdk.types import Logger


def bigquery_get_data(
    logger: Logger,
    dataset_id: str | None,
    table_id: str | None,
    big_query_hook: BigQueryHook,
    batch_size: int,
    selected_fields: list[str] | str | None,
) -> Iterator:
    logger.info("Fetching Data from:")
    logger.info("Dataset: %s ; Table: %s", dataset_id, table_id)

    for start_index in itertools.count(step=batch_size):
        rows: list[Row] | RowIterator = big_query_hook.list_rows(
            dataset_id=dataset_id,
            table_id=table_id,
            max_results=batch_size,
            selected_fields=selected_fields,
            start_index=start_index,
        )

        if isinstance(rows, RowIterator):
            raise TypeError("BigQueryHook.list_rows() returns iterator when return_iterator=False (default)")

        if len(rows) == 0:
            logger.info("Job Finished")
            return

        logger.info("Total Extracted rows: %s", len(rows) + start_index)

        yield [row.values() for row in rows]
